# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib import constants as n_consts
from neutron_lib.plugins import directory
from oslo_db import exception as db_exc
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.common import utils as n_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.scheduler import l3_ovn_scheduler as l3_sched
from neutron.tests.functional import base
from neutron.tests.functional.resources.ovsdb import events


class TestRouter(base.TestOVNFunctionalBase):
    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.chassis1 = self.add_fake_chassis(
            'ovs-host1', physical_nets=['physnet1', 'physnet3'],
            enable_chassis_as_gw=True, azs=[])
        self.chassis2 = self.add_fake_chassis(
            'ovs-host2', physical_nets=['physnet2', 'physnet3'],
            enable_chassis_as_gw=True, azs=[])
        self.physnet_used = ['physnet1', 'physnet2', 'physnet3']
        self.cr_lrp_pb_event = events.WaitForCrLrpPortBindingEvent()
        self.sb_api.idl.notify_handler.watch_event(self.cr_lrp_pb_event)

    def _create_router(self, name, gw_info=None, az_hints=None,
                       enable_ecmp=None, enable_bfd=None):
        router = {'router':
                  {'name': name,
                   'admin_state_up': True,
                   'tenant_id': self._tenant_id}}
        if az_hints:
            router['router']['availability_zone_hints'] = az_hints
        if gw_info:
            router['router']['external_gateway_info'] = gw_info
        if enable_bfd:
            router['router']['enable_default_route_bfd'] = enable_bfd
        if enable_ecmp:
            router['router']['enable_default_route_ecmp'] = enable_ecmp
        return self.l3_plugin.create_router(self.context, router)

    def _add_external_gateways(self, router_id, external_gateways):
        router = {'router': {'external_gateways': external_gateways}}
        return self.l3_plugin.add_external_gateways(
            self.context, router_id, body=router)

    def _remove_external_gateways(self, router_id, external_gateways):
        router = {'router': {'external_gateways': external_gateways}}
        return self.l3_plugin.remove_external_gateways(
            self.context, router_id, body=router)

    def _create_ext_network(self, name, net_type, physnet, seg,
                            gateway, cidr):
        arg_list = (pnet.NETWORK_TYPE, external_net.EXTERNAL,)
        net_arg = {pnet.NETWORK_TYPE: net_type,
                   external_net.EXTERNAL: True}
        if seg:
            arg_list = arg_list + (pnet.SEGMENTATION_ID,)
            net_arg[pnet.SEGMENTATION_ID] = seg
        if physnet:
            arg_list = arg_list + (pnet.PHYSICAL_NETWORK,)
            net_arg[pnet.PHYSICAL_NETWORK] = physnet
        network = self._make_network(self.fmt, name, True, as_admin=True,
                                     arg_list=arg_list, **net_arg)
        if cidr:
            self._make_subnet(self.fmt, network, gateway, cidr,
                              ip_version=n_consts.IP_VERSION_4)
        return network

    def _unset_lrp_gw_chassis(self, ovn_client):
        with ovn_client._nb_idl.transaction(check_error=True) as txn:
            for lrp in self.nb_api.tables['Logical_Router_Port'].rows.values():
                txn.add(ovn_client._nb_idl.update_lrouter_port(
                    lrp.name, gateway_chassis=[]))

    def _get_gwc_dict(self):
        sched_info = {}
        for row in self.nb_api.db_list_rows("Logical_Router_Port").execute(
                check_error=True):
            for gwc in row.gateway_chassis:
                chassis = sched_info.setdefault(gwc.chassis_name, {})
                chassis[gwc.priority] = chassis.get(gwc.priority, 0) + 1
        return sched_info

    def _create_routers_wait_pb(self, begin, n, gw_info=None,
                                bind_chassis=None):
        routers = []
        for i in range(begin, n + begin):
            try:
                router = self._create_router('router%d' % i, gw_info=gw_info)
                routers.append(router)
            except db_exc.DBReferenceError:
                # NOTE(ralonsoh): this is a workaround for LP#1956344. There
                # seems to be a bug in SQLite3. The "port" DB object is not
                # persistently stored in the DB and raises a "DBReferenceError"
                # exception occasionally.
                continue

            if bind_chassis:
                gw_port_id = router.get('gw_port_id')
                logical_port = 'cr-lrp-%s' % gw_port_id
                self.assertTrue(self.cr_lrp_pb_event.wait(logical_port),
                                msg='lrp %s failed to bind' % logical_port)
                self.sb_api.lsp_bind(logical_port, bind_chassis,
                                     may_exist=True).execute(check_error=True)
        return routers

    def _add_chassis(self, begin, n, physical_nets):
        chassis_added = []
        for i in range(begin, begin + n):
            chassis_added.append(
                self.add_fake_chassis(
                    'ovs-host%s' % i,
                    name=f'chassis-{i:02d}',
                    physical_nets=physical_nets,
                    other_config={
                        'ovn-cms-options': 'enable-chassis-as-gw'}))
        return chassis_added

    def test_gateway_chassis_on_router_gateway_port(self):
        ext2 = self._create_ext_network(
            'ext2', 'flat', 'physnet3', None, "20.0.0.1", "20.0.0.0/24")
        gw_info = {'network_id': ext2['network']['id']}
        self._create_router('router1', gw_info=gw_info)
        expected = [row.name for row in
                    self.sb_api.tables['Chassis'].rows.values()]
        for row in self.nb_api.tables[
                'Logical_Router_Port'].rows.values():
            chassis = [gwc.chassis_name for gwc in row.gateway_chassis]
            self.assertCountEqual(expected, chassis)

    def _check_gateway_chassis_candidates(self, candidates,
                                          router_az_hints=None,
                                          physnet='physnet1'):
        # In this test, fake_select() is called once from _create_router()
        # and later from schedule_unhosted_gateways()
        ovn_client = self.l3_plugin._ovn_client
        net_type = 'vlan' if physnet else 'geneve'
        ext1 = self._create_ext_network(
            'ext1', net_type, physnet, 1, "10.0.0.1", "10.0.0.0/24")
        # mock select function and check if it is called with expected
        # candidates.

        def fake_select(*args, **kwargs):
            self.assertCountEqual(candidates, kwargs['candidates'])
            # We are not interested in further processing, let us return
            # a random chassis name to avoid errors. If there are no
            # candidates, this method returns None.
            return ['a-random-chassis'] if candidates else None

        with mock.patch.object(self.l3_plugin.scheduler, 'select',
                               side_effect=fake_select) as plugin_select:
            gw_info = {'network_id': ext1['network']['id']}
            self._create_router('router1', gw_info=gw_info,
                                az_hints=router_az_hints)
            # If the network is tunnelled, the scheduler is not called.
            check = self.assertTrue if physnet else self.assertFalse
            check(plugin_select.called)
            plugin_select.reset_mock()

            # Unset the redirect-chassis so that schedule_unhosted_gateways
            # will try to schedule it.
            self._unset_lrp_gw_chassis(ovn_client)
            self.l3_plugin.schedule_unhosted_gateways()
            check = self.assertTrue if candidates else self.assertFalse
            check(plugin_select.called)

    def test_gateway_chassis_with_cms_and_bridge_mappings(self):
        # Both chassis1 and chassis3 are having proper bridge mappings,
        # but only chassis1 is having enable-chassis-as-gw.
        # Test if chassis1 is selected as candidate or not.
        self.chassis3 = self.add_fake_chassis(
            'ovs-host3', physical_nets=['physnet1'], azs=[])
        self._check_gateway_chassis_candidates([self.chassis1])

    def test_gateway_chassis_with_cms_and_no_bridge_mappings(self):
        # chassis1 is having proper bridge mappings.
        # chassis3 is having enable-chassis-as-gw, but no bridge mappings.
        self.chassis3 = self.add_fake_chassis(
            'ovs-host3',
            other_config={'ovn-cms-options': 'enable-chassis-as-gw'})
        ovn_client = self.l3_plugin._ovn_client
        ext1 = self._create_ext_network(
            'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
        # As we have 'gateways' in the system, but without required
        # chassis we should not schedule gw in that case at all.
        self._unset_lrp_gw_chassis(ovn_client)
        with mock.patch.object(self.l3_plugin.scheduler, 'select',
                               side_effect=[self.chassis1]):
            gw_info = {'network_id': ext1['network']['id']}
            self._create_router('router1', gw_info=gw_info)

        with mock.patch.object(
                ovn_client._nb_idl, 'update_lrouter_port') as ulrp:
            self.l3_plugin.schedule_unhosted_gateways()
            # Make sure that we don't schedule on chassis3
            # and do not updated the lrp port.
            ulrp.assert_not_called()

    def test_gateway_chassis_with_cms_and_azs(self):
        # Both chassis3 and chassis4 are having azs,
        # but only chassis3's azs is ['ovn'].
        # Test if chassis3 is selected as candidate or not.
        self.chassis3 = self.add_fake_chassis(
            'ovs-host3', physical_nets=['physnet1'],
            azs=['ovn'], enable_chassis_as_gw=True)
        self.chassis4 = self.add_fake_chassis(
            'ovs-host4', physical_nets=['physnet1'],
            azs=['ovn2'], enable_chassis_as_gw=True)
        self._check_gateway_chassis_candidates([self.chassis3],
                                               router_az_hints=['ovn'])

    def test_gateway_chassis_with_cms_and_not_match_azs(self):
        # add chassis3 is having azs [ovn], match router az_hints,
        # if not add chassis3, create router will fail with
        # AvailabilityZoneNotFound. after create will delete if.
        # add chassis4 is having azs [ovn2], not match routers az_hints [ovn]
        self.chassis3 = self.add_fake_chassis(
            'ovs-host3', physical_nets=['physnet1'], enable_chassis_as_gw=True)
        self.chassis4 = self.add_fake_chassis(
            'ovs-host4', physical_nets=['physnet1'], enable_chassis_as_gw=True,
            azs=['ovn2'])
        ovn_client = self.l3_plugin._ovn_client
        ext1 = self._create_ext_network(
            'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
        # As we have 'gateways' in the system, but without required
        # chassis we should not schedule gw in that case at all.
        self._unset_lrp_gw_chassis(ovn_client)
        with mock.patch.object(self.l3_plugin.scheduler, 'select',
                               side_effect=[self.chassis1]):
            gw_info = {'network_id': ext1['network']['id']}
            self._create_router('router1', gw_info=gw_info, az_hints=['ovn'])
        self.del_fake_chassis(self.chassis3)
        with mock.patch.object(
                ovn_client._nb_idl, 'update_lrouter_port') as ulrp:
            self.l3_plugin.schedule_unhosted_gateways()
            # Make sure that we don't schedule on chassis3
            # and do not updated the lrp port.
            ulrp.assert_not_called()

    def test_gateway_chassis_with_bridge_mappings_and_no_cms(self):
        # chassis1 is configured with proper bridge mappings,
        # but none of the chassis having enable-chassis-as-gw.
        # Test if chassis1 is selected as candidate or not.
        self._check_gateway_chassis_candidates([self.chassis1])

    def test_gateway_chassis_no_physnet_tunnelled_network(self):
        # The GW network is tunnelled, no physnet defined --> no possible
        # candidates.
        self._check_gateway_chassis_candidates(None, physnet=None)

    def test_gateway_chassis_least_loaded_scheduler(self):
        # This test will create 4 routers each with its own gateway.
        # Using the least loaded policy for scheduling gateway ports, we
        # expect that they are equally distributed across the two available
        # chassis.
        ovn_client = self.l3_plugin._ovn_client
        ovn_client._ovn_scheduler = l3_sched.OVNGatewayLeastLoadedScheduler()
        ext1 = self._create_ext_network(
            'ext1', 'flat', 'physnet3', None, "20.0.0.1", "20.0.0.0/24")
        gw_info = {'network_id': ext1['network']['id']}

        # Create 4 routers with a gateway. Since we're using physnet3, the
        # chassis candidates will be chassis1 and chassis2.
        for i in range(1, 5):
            self._create_router('router%d' % i, gw_info=gw_info)

        # At this point we expect two gateways to be present in chassis1
        # and two in chassis2. If schema supports L3 HA, we expect each
        # chassis to host 2 priority 2 gateways and 2 priority 1 ones.
        # Each chassis contains a dict of (priority, # of ports hosted).
        # {1: 2, 2: 2} means that this chassis hosts 2 ports of prio 1
        # and two ports of prio 2.
        expected = {self.chassis1: {1: 2, 2: 2},
                    self.chassis2: {1: 2, 2: 2}}
        sched_info = {}
        for row in self.nb_api.tables[
                'Logical_Router_Port'].rows.values():
            for gwc in row.gateway_chassis:
                chassis = sched_info.setdefault(gwc.chassis_name, {})
                chassis[gwc.priority] = chassis.get(gwc.priority, 0) + 1
        self.assertEqual(expected, sched_info)

    def test_gateway_chassis_least_loaded_scheduler_anti_affinity(self):
        ovn_client = self.l3_plugin._ovn_client
        ovn_client._ovn_scheduler = l3_sched.OVNGatewayLeastLoadedScheduler()
        ext1 = self._create_ext_network(
            'ext1', 'flat', 'physnet5', None, "10.10.50.1", "10.10.50.0/24")
        gw_info = {'network_id': ext1['network']['id']}

        chassis_list = []
        # first fill a few chassis with normal routers
        chassis_list.extend(
            self._add_chassis(0, ovn_const.MAX_GW_CHASSIS * 2, ['physnet5']))
        for i in range(1, (ovn_const.MAX_GW_CHASSIS * 4) + 1):
            router = self._create_router('router%d' % i, gw_info=gw_info)

        # add more chassis and create a set of routers with multiple gateway
        # ports
        #
        # This will stage a situation where a few chassis have higher load
        # which we can use to confirm that the anti-affinity algorithm works as
        # expected.
        #
        # Each router created below will have three LRPs, which should fit
        # in ovn_const.MAX_GW_CHASSIS * 3 chassis without duplicates when
        # using the anti affinity scheduler.
        chassis_list.extend(
            self._add_chassis(
                len(chassis_list), ovn_const.MAX_GW_CHASSIS, ['physnet5']))
        router_lrps = {}
        for i in range(1, 2 + 1):
            router = self._create_router('router-multi-gw%d' % i)
            router_lrps[router['id']] = []
            self._add_external_gateways(
                router['id'],
                [
                    {'network_id': ext1['network']['id']},
                    {'network_id': ext1['network']['id']},
                    {'network_id': ext1['network']['id']},
                ])
            for row in self.nb_api.tables[
                    'Logical_Router_Port'].rows.values():
                if (ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY
                        not in row.external_ids):
                    continue
                if row.external_ids[
                        ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY] == router['id']:
                    chassis = {}
                    for gwc in row.gateway_chassis:
                        chassis[gwc.priority] = gwc.chassis_name
                    router_lrps[router['id']].append(chassis)
        for router, lrp_lists in router_lrps.items():
            while lrp_lists:
                try:
                    lrps = lrp_lists.pop()
                except IndexError:
                    break
                for lrp_list in lrp_lists:
                    for n in range(1, ovn_const.MAX_GW_CHASSIS + 1):
                        self.assertNotEqual(
                            lrps[n],
                            lrp_list[n])

    def _get_gw_port(self, router_id):
        router = self.l3_plugin._get_router(self.context, router_id)
        gw_port_id = router.get('gw_port_id', '')
        for row in self.nb_api.tables['Logical_Router_Port'].rows.values():
            if row.name == 'lrp-%s' % gw_port_id:
                return row

    def test_gateway_chassis_with_subnet_changes(self):
        """Launchpad bug #1843485: logical router port is getting lost

        Test cases when subnets are added to an external network after router
        has been configured to use that network via "set --external-gateway"
        """
        with mock.patch.object(self.l3_plugin.scheduler, 'select',
                               return_value=self.chassis1) as plugin_select:
            router1 = self._create_router('router1', gw_info=None)
            router_id = router1['id']
            self.assertIsNone(self._get_gw_port(router_id),
                              "router logical port unexpected before ext net")

            # Create external network with no subnets and assign it to router
            ext1 = self._create_ext_network(
                'ext1', 'flat', 'physnet3', None, gateway=None, cidr=None)
            net_id = ext1['network']['id']

            gw_info = {'network_id': ext1['network']['id']}
            self.l3_plugin.update_router(
                self.context, router_id,
                {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
            self.assertIsNotNone(self._get_gw_port(router_id),
                                 "router logical port must exist after gw add")

            # Add subnets to external network. This should percolate
            # into l3_plugin.update_router()
            kwargs = {'ip_version': n_consts.IP_VERSION_4,
                      'gateway_ip': '10.0.0.1', 'cidr': '10.0.0.0/24'}
            subnet4_res = self._create_subnet(
                self.fmt, net_id, **kwargs)
            subnet4 = self.deserialize(self.fmt, subnet4_res).get('subnet')
            self.assertIsNotNone(self._get_gw_port(router_id),
                                 "router logical port must exist after v4 add")

            kwargs = {'ip_version': n_consts.IP_VERSION_6,
                      'gateway_ip': 'fe81::1', 'cidr': 'fe81::/64',
                      'ipv6_ra_mode': n_consts.IPV6_SLAAC,
                      'ipv6_address_mode': n_consts.IPV6_SLAAC}
            subnet6_res = self._create_subnet(
                self.fmt, net_id, **kwargs)
            subnet6 = self.deserialize(self.fmt, subnet6_res).get('subnet')
            self.assertIsNotNone(self._get_gw_port(router_id),
                                 "router logical port must exist after v6 add")

            self.assertGreaterEqual(plugin_select.call_count, 3)

            # Verify that ports have had the subnets created
            kwargs = {'device_owner': n_consts.DEVICE_OWNER_ROUTER_GW}
            ports_res = self._list_ports(self.fmt, net_id=net_id, **kwargs)
            ports = self.deserialize(self.fmt, ports_res).get('ports')
            subnet4_ip = None
            subnet6_ip = None
            for port in ports:
                for fixed_ip in port.get('fixed_ips', []):
                    if fixed_ip.get('subnet_id') == subnet4['id']:
                        subnet4_ip = fixed_ip.get('ip_address')
                    if fixed_ip.get('subnet_id') == subnet6['id']:
                        subnet6_ip = fixed_ip.get('ip_address')
            self.assertIsNotNone(subnet4_ip)
            self.assertIsNotNone(subnet6_ip)

            # Verify that logical router port is properly configured
            gw_port = self._get_gw_port(router_id)
            self.assertIsNotNone(gw_port)

            expected_networks = ['%s/24' % subnet4_ip, '%s/64' % subnet6_ip]
            self.assertCountEqual(
                expected_networks, gw_port.networks,
                'networks in ovn port must match fixed_ips in neutron')

    def test_logical_router_port_creation(self):
        """Launchpad bug #1844652: Verify creation and removal of lrp

        This test verifies that logical router port is created and removed
        based on attaching and detaching the external network to a router.
        """
        router = self._create_router('router1', gw_info=None)
        router_id = router['id']
        self.assertIsNone(self._get_gw_port(router_id),
                          "router logical port unexpected before ext net")

        # Create external network and assign it to router
        ext1 = self._create_ext_network(
            'ext1', 'flat', 'physnet3', None, gateway=None, cidr=None)
        gw_info = {'network_id': ext1['network']['id']}
        self.l3_plugin.update_router(
            self.context, router_id,
            {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
        self.assertIsNotNone(self._get_gw_port(router_id),
                             "router logical port missing after ext net add")

        # Un-assign external network from router
        self.l3_plugin.update_router(
            self.context, router_id,
            {'router': {l3_apidef.EXTERNAL_GW_INFO: None}})
        self.assertIsNone(self._get_gw_port(router_id),
                          "router logical port exists after ext net removal")

    def test_gateway_chassis_with_bridge_mappings(self):
        """Check selected ovn chassis based on external network

        This test sets different gateway values to ensure that the proper
        chassis are candidates, based on the physical network mappings.
        """

        ovn_client = self.l3_plugin._ovn_client
        # Create external networks with vlan, flat and geneve network types
        ext1 = self._create_ext_network(
            'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
        ext2 = self._create_ext_network(
            'ext2', 'flat', 'physnet3', None, "20.0.0.1", "20.0.0.0/24")
        ext3 = self._create_ext_network(
            'ext3', 'geneve', None, 10, "30.0.0.1", "30.0.0.0/24")
        # mock select function and check if it is called with expected
        # candidates.
        self.candidates = []

        def fake_select(*args, **kwargs):
            self.assertCountEqual(self.candidates, kwargs['candidates'])
            # We are not interested in further processing, let us return
            # a random chassis name to avoid errors.
            return ['a-random-chassis']

        with mock.patch.object(self.l3_plugin.scheduler, 'select',
                               side_effect=fake_select) as plugin_select:
            self.candidates = [self.chassis1]
            gw_info = {'network_id': ext1['network']['id']}
            router1 = self._create_router('router1', gw_info=gw_info)

            # Unset the redirect-chassis so that schedule_unhosted_gateways
            # will try to schedule it.
            self._unset_lrp_gw_chassis(ovn_client)
            self.l3_plugin.schedule_unhosted_gateways()

            self.candidates = [self.chassis1, self.chassis2]
            gw_info = {'network_id': ext2['network']['id']}
            self.l3_plugin.update_router(
                self.context, router1['id'],
                {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
            self._unset_lrp_gw_chassis(ovn_client)
            self.l3_plugin.schedule_unhosted_gateways()

            self.candidates = []
            gw_info = {'network_id': ext3['network']['id']}
            self.l3_plugin.update_router(
                self.context, router1['id'],
                {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
            self._unset_lrp_gw_chassis(ovn_client)
            self.l3_plugin.schedule_unhosted_gateways()

            # We can't test call_count for these mocks, as we have disabled
            # maintenance_worker which will trigger chassis events
            # and eventually calling schedule_unhosted_gateways.
            # The router is created with a gateway port and updated twice.
            # However, the "plugin_select" is called only twice because the
            # third gateway network used is type "geneve" and the LRP are not
            # hosted in any chassis.
            self.assertGreaterEqual(plugin_select.call_count, 2)

    def _find_port_binding(self, port_id):
        cmd = self.sb_api.db_find_rows('Port_Binding',
                                       ('logical_port', '=', port_id))
        rows = cmd.execute(check_error=True)
        return rows[0] if rows else None

    def test_router_gateway_port_binding_host_id(self):
        # Test setting chassis on chassisredirect port in Port_Binding table,
        # will update host_id of corresponding router gateway port
        # with this chassis.
        chassis = idlutils.row_by_value(self.sb_api.idl, 'Chassis',
                                        'name', self.chassis1)
        host_id = chassis.hostname
        ext = self._create_ext_network(
            'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
        gw_info = {'network_id': ext['network']['id']}
        router = self._create_router('router1', gw_info=gw_info)
        core_plugin = directory.get_plugin()
        gw_port_id = router.get('gw_port_id')

        # Set chassis on chassisredirect port in Port_Binding table
        logical_port = 'cr-lrp-%s' % gw_port_id
        self.assertTrue(self.cr_lrp_pb_event.wait(logical_port),
                        msg='lrp %s failed to bind' % logical_port)
        self.sb_api.lsp_bind(logical_port, self.chassis1,
                             may_exist=True).execute(check_error=True)

        def check_port_binding_host_id(port_id):
            # Get port from Neutron DB
            port = core_plugin.get_ports(
                self.context, filters={'id': [port_id]})[0]
            # Get port from OVN DB
            bp = self._find_port_binding(port_id)
            ovn_host_id = bp.external_ids.get(ovn_const.OVN_HOST_ID_EXT_ID_KEY)
            return port[portbindings.HOST_ID] == host_id == ovn_host_id

        # Test if router gateway port updated with this chassis
        n_utils.wait_until_true(lambda: check_port_binding_host_id(
            gw_port_id))

        # Simulate failover to another chassis and check host_id in Neutron DB
        # and external_ids:neutron:host_id in OVN DB are updated
        chassis = idlutils.row_by_value(
            self.sb_api.idl, "Chassis", "name", self.chassis2
        )
        host_id = chassis.hostname
        self.sb_api.lsp_unbind(logical_port).execute(check_error=True)
        self.sb_api.lsp_bind(logical_port, self.chassis2).execute(
            check_error=True
        )
        n_utils.wait_until_true(lambda: check_port_binding_host_id(gw_port_id))

    def _validate_router_ipv6_ra_configs(self, lrp_name, expected_ra_confs):
        lrp = idlutils.row_by_value(self.nb_api.idl,
                                    'Logical_Router_Port', 'name', lrp_name)
        self.assertEqual(expected_ra_confs, lrp.ipv6_ra_configs)

    def _test_router_port_ipv6_ra_configs_helper(
            self, cidr='aef0::/64', ip_version=6,
            address_mode=n_consts.IPV6_SLAAC,):
        router1 = self._create_router('router1')
        n1 = self._make_network(self.fmt, 'n1', True)
        if ip_version == 6:
            kwargs = {'ip_version': 6, 'cidr': 'aef0::/64',
                      'ipv6_address_mode': address_mode,
                      'ipv6_ra_mode': address_mode}
        else:
            kwargs = {'ip_version': 4, 'cidr': '10.0.0.0/24'}

        res = self._create_subnet(self.fmt, n1['network']['id'],
                                  **kwargs)

        n1_s1 = self.deserialize(self.fmt, res)
        n1_s1_id = n1_s1['subnet']['id']
        router_iface_info = self.l3_plugin.add_router_interface(
            self.context, router1['id'], {'subnet_id': n1_s1_id})

        lrp_name = ovn_utils.ovn_lrouter_port_name(
            router_iface_info['port_id'])
        if ip_version == 6:
            expected_ra_configs = {
                'address_mode': ovn_utils.get_ovn_ipv6_address_mode(
                    address_mode),
                'send_periodic': 'true',
                'mtu': '1442'}
        else:
            expected_ra_configs = {}
        self._validate_router_ipv6_ra_configs(lrp_name, expected_ra_configs)

    def test_router_port_ipv6_ra_configs_addr_mode_slaac(self):
        self._test_router_port_ipv6_ra_configs_helper()

    def test_router_port_ipv6_ra_configs_addr_mode_dhcpv6_stateful(self):
        self._test_router_port_ipv6_ra_configs_helper(
            address_mode=n_consts.DHCPV6_STATEFUL)

    def test_router_port_ipv6_ra_configs_addr_mode_dhcpv6_stateless(self):
        self._test_router_port_ipv6_ra_configs_helper(
            address_mode=n_consts.DHCPV6_STATELESS)

    def test_router_port_ipv6_ra_configs_ipv4(self):
        self._test_router_port_ipv6_ra_configs_helper(
            ip_version=4)

    def test_create_delete_router_multiple_gw_ports(self):
        ext4 = self._create_ext_network(
            'ext4', 'flat', self.physnet_used[0], None, '40.0.0.1',
            '40.0.0.0/24')
        router = self._create_router('router4')
        gws = self._add_external_gateways(
            router['id'],
            [
                {'network_id': ext4['network']['id']},
                {'network_id': ext4['network']['id']},
            ]
        )
        lr = self.nb_api.lookup('Logical_Router',
                                ovn_utils.ovn_name(router['id']))
        self.assertEqual(
            len(lr.ports),
            len(gws['router']['external_gateways']))

        self.assertEqual(
            len(lr.static_routes),
            1)

        self.l3_plugin.delete_router(self.context, id=router['id'])
        self.assertRaises(idlutils.RowNotFound, self.nb_api.lookup,
                          'Logical_Router', ovn_utils.ovn_name(router['id']))

    def test_create_router_multiple_gw_ports_ecmp(self):
        ext5 = self._create_ext_network(
            'ext5', 'flat', self.physnet_used[1], None, '10.0.50.1',
            '10.0.50.0/24')
        router = self._create_router('router5', enable_ecmp=True)
        gws = self._add_external_gateways(
            router['id'],
            [
                {'network_id': ext5['network']['id']},
                {'network_id': ext5['network']['id']},
            ]
        )
        lr = self.nb_api.lookup('Logical_Router',
                                ovn_utils.ovn_name(router['id']))
        # Check that the expected number of ports are created
        self.assertEqual(
            len(lr.ports),
            len(gws['router']['external_gateways']))
        # Check that the expected number of static routes are created
        self.assertEqual(
            len(lr.static_routes),
            len(gws['router']['external_gateways']))

    def test_create_delete_router_multiple_gw_ports_ecmp_and_bfd(self):
        default_gw = "10.0.60.1"
        ext6 = self._create_ext_network(
            'ext6', 'flat', 'physnet6', None, default_gw, "10.0.60.0/24")
        router = self._create_router('router6', gw_info=None,
                                     enable_bfd=True, enable_ecmp=True)
        gws = self._add_external_gateways(
            router['id'],
            [
                {'network_id': ext6['network']['id']},
                {'network_id': ext6['network']['id']},
            ])
        lr = self.nb_api.lookup('Logical_Router',
                                ovn_utils.ovn_name(router['id']))

        # Check that the expected number of ports are created
        self.assertEqual(
            len(lr.ports),
            len(gws['router']['external_gateways']))
        # Check that the expected number of static routes are created
        self.assertEqual(
            len(lr.static_routes),
            len(gws['router']['external_gateways']))
        # Check that static_route bfd and output_port attributes is set to the
        # expected values
        for static_route in lr.static_routes:
            self.assertNotEqual(
                [],
                static_route.bfd)
            self.assertNotEqual(
                [],
                static_route.output_port)
            self.assertIn(static_route.output_port[0],
                          [lrp.name
                           for lrp in lr.ports])
            self.assertIn(static_route.bfd[0].logical_port,
                          [lrp.name
                           for lrp in lr.ports])
            self.assertEqual(static_route.bfd[0].logical_port,
                             static_route.output_port[0])

        router_ips = set()
        for ext_gws in gws['router']['external_gateways']:
            for ext_fip in ext_gws['external_fixed_ips']:
                router_ips.add(ext_fip['ip_address'])

        lrps = set()
        for lrp in lr.ports:
            for network in lrp.networks:
                self.assertIn(
                    network.split('/')[0],
                    router_ips)
            lrps.add(lrp.name)
            bfd_rows = self.nb_api.bfd_find(
                    lrp.name, default_gw).execute(check_error=True)
            if not bfd_rows:
                raise AssertionError('None of the expected BFD rows found.')
            for bfd_row in bfd_rows:
                self.assertEqual(
                    bfd_row.logical_port,
                    lrp.name)
                self.assertEqual(
                    bfd_row.dst_ip,
                    default_gw)

        self.l3_plugin.delete_router(self.context, id=router['id'])
        self.assertRaises(idlutils.RowNotFound, self.nb_api.lookup,
                          'Logical_Router', ovn_utils.ovn_name(router['id']))
        for lrp_name in lrps:
            if self.nb_api.bfd_find(
                    lrp.name, default_gw).execute(check_error=True):
                raise AssertionError('Unexpectedly found BFD rows.')

    def test_update_router_single_gw_bfd(self):
        ext1 = self._create_ext_network(
            'ext7', 'flat', 'physnet1', None, "10.0.70.1", "10.0.70.0/24")
        gw_info = {'network_id': ext1['network']['id']}
        router = self._create_router('router7', gw_info=gw_info)
        self.assertFalse(router['enable_default_route_bfd'])
        lr = self.nb_api.lr_get(ovn_utils.ovn_name(router['id'])).execute()
        for route in ovn_utils.get_lrouter_ext_gw_static_route(lr):
            self.assertEqual(
                [],
                route.bfd)

        router = self.l3_plugin.update_router(
            self.context, router['id'],
            {'router': {'enable_default_route_bfd': True}})
        self.assertTrue(router['enable_default_route_bfd'])
        lr = self.nb_api.lr_get(ovn_utils.ovn_name(router['id'])).execute()
        for route in ovn_utils.get_lrouter_ext_gw_static_route(lr):
            self.assertNotEqual(
                [],
                route.bfd)

    def test_gateway_chassis_rebalance(self):
        ovn_client = self.l3_plugin._ovn_client
        chassis4 = self.add_fake_chassis(
            'ovs-host4', physical_nets=['physnet4'], other_config={
                'ovn-cms-options': 'enable-chassis-as-gw'})
        ovn_client._ovn_scheduler = l3_sched.OVNGatewayLeastLoadedScheduler()
        ext1 = self._create_ext_network(
            'ext1', 'flat', 'physnet4', None, "30.0.0.1", "30.0.0.0/24")
        gw_info = {'network_id': ext1['network']['id']}
        # Tries to create 5 routers with a gateway. Since we're using
        # physnet4, the chassis candidates will be chassis4 initially.
        num_routers = len(self._create_routers_wait_pb(
            1, 20, gw_info=gw_info, bind_chassis=chassis4))
        self.l3_plugin.schedule_unhosted_gateways()
        expected = {chassis4: {1: num_routers}}
        self.assertEqual(expected, self._get_gwc_dict())

        # Add another chassis as a gateway chassis
        chassis5 = self.add_fake_chassis(
            'ovs-host5', physical_nets=['physnet4'], other_config={
                'ovn-cms-options': 'enable-chassis-as-gw'})
        # Add a node as compute node. Compute node wont be
        # used to schedule the router gateway ports therefore
        # priority values wont be changed. Therefore chassis4 would
        # still have priority 2
        self.add_fake_chassis('ovs-host6', physical_nets=['physnet4'])

        # Chassis4 should have all ports at Priority 2
        self.l3_plugin.schedule_unhosted_gateways()
        self.assertEqual({2: num_routers}, self._get_gwc_dict()[chassis4])
        # Chassis5 should have all ports at Priority 1
        self.assertEqual({1: num_routers}, self._get_gwc_dict()[chassis5])

        # delete chassis that hosts all the gateways
        self.del_fake_chassis(chassis4)
        self.l3_plugin.schedule_unhosted_gateways()

        # As Chassis4 has been removed so all gateways that were
        # hosted there are now primaries on chassis5 and have
        # priority 1.
        self.assertEqual({1: num_routers}, self._get_gwc_dict()[chassis5])

    def test_gateway_chassis_rebalance_max_chassis(self):
        chassis_list = []
        # spawn 6 chassis and check if port has MAX_CHASSIS candidates.
        for i in range(0, ovn_const.MAX_GW_CHASSIS + 1):
            chassis_list.append(
                self.add_fake_chassis(
                    'ovs-host%s' % i, physical_nets=['physnet1'],
                    other_config={'ovn-cms-options': 'enable-chassis-as-gw'}))

        ext1 = self._create_ext_network(
            'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
        gw_info = {'network_id': ext1['network']['id']}
        router = self._create_router('router', gw_info=gw_info)
        gw_port_id = router.get('gw_port_id')
        logical_port = 'cr-lrp-%s' % gw_port_id
        self.assertTrue(self.cr_lrp_pb_event.wait(logical_port),
                        msg='lrp %s failed to bind' % logical_port)
        self.sb_api.lsp_bind(logical_port, chassis_list[0],
                             may_exist=True).execute(check_error=True)

        self.l3_plugin.schedule_unhosted_gateways()
        for row in self.nb_api.tables[
                'Logical_Router_Port'].rows.values():
            self.assertEqual(ovn_const.MAX_GW_CHASSIS,
                             len(row.gateway_chassis))

    def test_set_router_mac_age_limit(self):
        name = "macage_router1"
        router = self._create_router(name)
        lr_name = ovn_utils.ovn_name(router['id'])
        options = self.nb_api.db_get(
            "Logical_Router", lr_name, "options").execute(check_error=True)
        self.assertEqual(ovn_conf.get_ovn_mac_binding_age_threshold(),
                         options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT])

    def test_schedule_unhosted_gateways_single_transaction(self):
        ext1 = self._create_ext_network(
            'ext1', 'flat', 'physnet6', None, "10.0.60.1", "10.0.60.0/24")
        gw_info = {'network_id': ext1['network']['id']}

        # Attempt to add 4 routers, since there are no chassis, none of them
        # will be scheduled on any chassis.
        num_routers = len(self._create_routers_wait_pb(1, 4, gw_info=gw_info))
        self.assertEqual({}, self._get_gwc_dict())

        # Add 2 chassis and rebalance gateways.
        #
        # The ovsdb_monitor.ChassisEvent handler will attempt to schedule
        # unhosted gateways as chassis are added.
        #
        # Temporarily mock it out while adding the chassis so that we can get a
        # predictable result for the purpose of this test.
        chassis_list = []
        with mock.patch.object(
                self.l3_plugin, 'schedule_unhosted_gateways'):
            chassis_list.extend(self._add_chassis(1, 2, ['physnet6']))
        self.assertEqual({}, self._get_gwc_dict())

        # Wrap `self.l3_plugin._nb_ovn.transaction` so that we can assert on
        # number of calls.
        with mock.patch.object(
                self.l3_plugin._nb_ovn, 'transaction',
                wraps=self.l3_plugin._nb_ovn.transaction) as wrap_txn:
            self.l3_plugin.schedule_unhosted_gateways()
            # The server is alive and we can't control the exact number of
            # calls made to `_nb_ovn.transaction`. We can however check that
            # the number of calls is less than number of unhosted gateways.
            self.assertLess(len(wrap_txn.mock_calls), num_routers)

        # Ensure the added gateways are spread evenly on the added chassis.
        self.assertEqual(
            {'chassis-01': {1: 2, 2: 2},
             'chassis-02': {1: 2, 2: 2}},
            self._get_gwc_dict())



# Copyright (c) 2024 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing

from oslo_utils import timeutils

from neutron.common import utils


def get_start_time(default=None, current_time=False):
    """Return the 'start-time=%t' config varible in the WSGI config

    This variable contains the start time of the WSGI server. Check
    https://uwsgi-docs.readthedocs.io/en/latest/Configuration.html
    #magic-variables

    :param default: (int or float) in case the uwsgi option 'start-time' is not
                    available or the uwsgi module cannot be loaded, the method
                    will return this value.
    :param current_time: (bool) if ``default`` is None and this flag is set,
                         the method will return the current time.
    :return: (int) start time in seconds.
    """
    if not default and current_time:
        default = utils.datetime_to_ts(timeutils.utcnow())
    default = int(default) if default else None
    try:
        # pylint: disable=import-outside-toplevel
        import uwsgi
        start_time = uwsgi.opt.get('start-time')
        if not start_time:
            return default
        return int(start_time.decode(encoding='utf-8'))
    except ImportError:
        return default


def get_api_worker_id() -> typing.Union[int, None]:
    """Return the worker ID number provided by uWSGI"""
    try:
        # pylint: disable=import-outside-toplevel
        import uwsgi
        return uwsgi.worker_id()
    except ImportError:
        return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
from ovsdbapp.backend.ovs_idl import vlog

from neutron._i18n import _
from neutron.common.ovn import constants as ovn_const
from neutron.conf.agent import ovs_conf

LOG = logging.getLogger(__name__)

EXTRA_LOG_LEVEL_DEFAULTS = [
]

VLOG_LEVELS = {'CRITICAL': vlog.CRITICAL, 'ERROR': vlog.ERROR, 'WARNING':
               vlog.WARN, 'INFO': vlog.INFO, 'DEBUG': vlog.DEBUG}

OVN_NB_GLOBAL = "ovn_nb_global"

ovn_opts = [
    cfg.ListOpt('ovn_nb_connection',
                default=['tcp:127.0.0.1:6641'],
                item_type=types.String(regex=r'^(tcp|ssl|unix):.+'),
                help=_('The connection string for the OVN_Northbound OVSDB.\n'
                       'Use tcp:IP:PORT for TCP connection.\n'
                       'Use ssl:IP:PORT for SSL connection. The '
                       'ovn_nb_private_key, ovn_nb_certificate and '
                       'ovn_nb_ca_cert are mandatory.\n'
                       'Use unix:FILE for unix domain socket connection.\n'
                       'Multiple connections can be specified by a comma '
                       'separated string. See also: '
                       'https://github.com/openvswitch/ovs/blob'
                       '/ab4d3bfbef37c31331db5a9dbe7c22eb8d5e5e5f'
                       '/python/ovs/db/idl.py#L215-L216')),
    cfg.StrOpt('ovn_nb_private_key',
               default='',
               help=_('The PEM file with private key for SSL connection to '
                      'OVN-NB-DB')),
    cfg.StrOpt('ovn_nb_certificate',
               default='',
               help=_('The PEM file with certificate that certifies the '
                      'private key specified in ovn_nb_private_key')),
    cfg.StrOpt('ovn_nb_ca_cert',
               default='',
               help=_('The PEM file with CA certificate that OVN should use to'
                      ' verify certificates presented to it by SSL peers')),
    cfg.ListOpt('ovn_sb_connection',
                default=['tcp:127.0.0.1:6642'],
                item_type=types.String(regex=r'^(tcp|ssl|unix):.+'),
                help=_('The connection string for the OVN_Southbound OVSDB.\n'
                       'Use tcp:IP:PORT for TCP connection.\n'
                       'Use ssl:IP:PORT for SSL connection. The '
                       'ovn_sb_private_key, ovn_sb_certificate and '
                       'ovn_sb_ca_cert are mandatory.\n'
                       'Use unix:FILE for unix domain socket connection.\n'
                       'Multiple connections can be specified by a comma '
                       'separated string. See also: '
                       'https://github.com/openvswitch/ovs/blob'
                       '/ab4d3bfbef37c31331db5a9dbe7c22eb8d5e5e5f'
                       '/python/ovs/db/idl.py#L215-L216')),
    cfg.StrOpt('ovn_sb_private_key',
               default='',
               help=_('The PEM file with private key for SSL connection to '
                      'OVN-SB-DB')),
    cfg.StrOpt('ovn_sb_certificate',
               default='',
               help=_('The PEM file with certificate that certifies the '
                      'private key specified in ovn_sb_private_key')),
    cfg.StrOpt('ovn_sb_ca_cert',
               default='',
               help=_('The PEM file with CA certificate that OVN should use to'
                      ' verify certificates presented to it by SSL peers')),
    cfg.IntOpt('ovsdb_connection_timeout',
               default=180,
               help=_('Timeout, in seconds, for the OVSDB '
                      'connection transaction')),
    cfg.IntOpt('ovsdb_retry_max_interval',
               default=180,
               help=_('Max interval, in seconds ,between '
                      'each retry to get the OVN NB and SB IDLs')),
    cfg.IntOpt('ovsdb_probe_interval',
               min=0,
               default=60000,
               help=_('The probe interval for the OVSDB session, in '
                      'milliseconds. If this is zero, it disables the '
                      'connection keepalive feature. If non-zero the value '
                      'will be forced to at least 1000 milliseconds. Defaults '
                      'to 60 seconds.')),
    cfg.StrOpt('neutron_sync_mode',
               default=ovn_const.OVN_DB_SYNC_MODE_LOG,
               choices=[(ovn_const.OVN_DB_SYNC_MODE_OFF,
                         "Synchronization is off."),
                        (ovn_const.OVN_DB_SYNC_MODE_LOG,
                         "During neutron-server startup, check to see if OVN "
                         "is in sync with the Neutron database. "
                         "Log warnings for any inconsistencies found so that "
                         "an admin can investigate."),
                        (ovn_const.OVN_DB_SYNC_MODE_REPAIR,
                         "During neutron-server startup, automatically create "
                         "resources found in Neutron but not in OVN. "
                         "Also remove resources from OVN that are no longer "
                         "found in Neutron."),
                        (ovn_const.OVN_DB_SYNC_MODE_MIGRATE,
                         "This mode is to OVS to OVN migration. "
                         "It will sync the DB just like repair mode but it "
                         "will additionally fix the Neutron DB resource from "
                         "OVS to OVN.")],
               help=_('The synchronization mode of OVN_Northbound OVSDB '
                      'with Neutron DB.')),
    cfg.StrOpt("ovn_l3_scheduler",
               default=ovn_const.OVN_L3_SCHEDULER_LEASTLOADED,
               choices=[(ovn_const.OVN_L3_SCHEDULER_LEASTLOADED,
                         "Select chassis with fewest gateway ports."),
                        (ovn_const.OVN_L3_SCHEDULER_CHANCE,
                         "Select chassis randomly.")],
               help=_('The OVN L3 Scheduler type used to schedule router '
                      'gateway ports on hypervisors/chassis.')),
    cfg.BoolOpt('enable_distributed_floating_ip',
                default=False,
                help=_('Enable distributed floating IP support.\n'
                       'If True, the NAT action for floating IPs will be done '
                       'locally and not in the centralized gateway. This '
                       'saves the path to the external network. This requires '
                       'the user to configure the physical network map '
                       '(i.e. ovn-bridge-mappings) on each compute node.')),
    cfg.StrOpt("vhost_sock_dir",
               default="/var/run/openvswitch",
               help=_("The directory in which vhost virtio sockets "
                      "are created by all the vswitch daemons")),
    cfg.IntOpt('dhcp_default_lease_time',
               default=(12 * 60 * 60),
               help=_('Default lease time (in seconds) to use with '
                      'OVN\'s native DHCP service.')),
    cfg.StrOpt("ovsdb_log_level",
               default="INFO",
               choices=list(VLOG_LEVELS.keys()),
               help=_("The log level used for OVSDB")),
    cfg.BoolOpt('ovn_metadata_enabled',
                default=False,
                help=_('Whether to use metadata service.')),
    cfg.ListOpt('dns_servers',
                default=[],
                help=_("Comma-separated list of the DNS servers which will be "
                       "used as forwarders if a subnet's dns_nameservers "
                       "field is empty. If both subnet's dns_nameservers and "
                       "this option are empty, then the DNS resolvers on the "
                       "host running the neutron server will be used.")),
    cfg.BoolOpt('dns_records_ovn_owned',
                default=False,
                help=_("Whether to consider DNS records local to OVN or not. "
                       "For OVN version 24.03 and above if this option is set "
                       "to True, DNS records will be treated local to the OVN "
                       "controller and it will respond to the queries for the "
                       "records and record types known to it, else it will "
                       "forward them to the configured DNS server(s).")),
    cfg.DictOpt('ovn_dhcp4_global_options',
                default={},
                help=_("Dictionary of global DHCPv4 options which will be "
                       "automatically set on each subnet upon creation and "
                       "on all existing subnets when Neutron starts.\n"
                       "An empty value for a DHCP option will cause that "
                       "option to be unset globally.\n"
                       "EXAMPLES:\n"
                       "- ntp_server:1.2.3.4,wpad:1.2.3.5 - Set ntp_server "
                       "and wpad\n"
                       "- ntp_server:,wpad:1.2.3.5 - Unset ntp_server and "
                       "set wpad\n"
                       "See the ovn-nb(5) man page for available options.")),
    cfg.DictOpt('ovn_dhcp6_global_options',
                default={},
                help=_("Dictionary of global DHCPv6 options which will be "
                       "automatically set on each subnet upon creation and "
                       "on all existing subnets when Neutron starts.\n"
                       "An empty value for a DHCPv6 option will cause that "
                       "option to be unset globally.\n"
                       "See the ovn-nb(5) man page for available options.")),
    cfg.BoolOpt('ovn_emit_need_to_frag',
                default=True,
                deprecated_for_removal=True,
                deprecated_since='2025.1',
                deprecated_reason=("The option is useful only on very old "
                                   "Linux kernels (version < 5.2)."),
                help=_('Configure OVN to emit "need to frag" packets in '
                       'case of MTU mismatches.\n'
                       'You may have to disable this option if you are '
                       'running an old host kernel (version < 5.2). You may '
                       'check the output of the following command:\n'
                       'ovs-appctl -t ovs-vswitchd dpif/show-dp-features '
                       'br-int | grep "Check pkt length action".')),
    cfg.BoolOpt('disable_ovn_dhcp_for_baremetal_ports',
                default=False,
                help=_('Disable OVN\'s built-in DHCP for baremetal ports '
                       '(VNIC type "baremetal"). This allows operators to '
                       'plug their own DHCP server of choice for PXE booting '
                       'baremetal nodes. OVN 23.06.0 and newer also supports '
                       'baremetal ``PXE`` based provisioning over IPv6. '
                       'If an older version of OVN is used for baremetal '
                       'provisioning over IPv6 this option should be set '
                       'to "True" and neutron-dhcp-agent should be used '
                       'instead. Defaults to "False".')),
    cfg.BoolOpt('localnet_learn_fdb',
                default=False,
                help=_('If enabled it will allow localnet ports to learn MAC '
                       'addresses and store them in FDB SB table. This avoids '
                       'flooding for traffic towards unknown IPs when port '
                       'security is disabled. It requires OVN 22.09 or '
                       'newer.')),
    cfg.IntOpt('fdb_age_threshold',
               min=0,
               default=0,
               help=_('The number of seconds to keep FDB entries in the OVN '
                      'DB. The value defaults to 0, which means disabled. '
                      'This is supported by OVN >= 23.09.')),
    cfg.IntOpt('mac_binding_age_threshold',
               min=0,
               default=0,
               help=_('The number of seconds to keep MAC_Binding entries in '
                      'the OVN DB. 0 to disable aging.')),
    cfg.BoolOpt('broadcast_arps_to_all_routers',
                default=True,
                help=_('If enabled (default) OVN will flood ARP requests to '
                       'all attached ports on a network. If set to False, '
                       'ARP requests are only sent to routers on that network '
                       'if the target MAC address matches. ARP requests that '
                       'do not match a router will only be forwarded to '
                       'non-router ports. Supported by OVN >= 23.06.')),
    cfg.BoolOpt('ovn_router_indirect_snat',
                default=False,
                help=_('Whether to configure SNAT for all nested subnets '
                       'connected to the router through any other routers, '
                       'similar to the default ML2/OVS behavior. Defaults to '
                       '"False".')),
    cfg.StrOpt('live_migration_activation_strategy',
               default="rarp",
               choices=[("rarp",
                         "Expect the hypervisor to send a Reverse ARP request "
                         "through the migrated port after migration is "
                         "complete."),
                        ("",
                         "A migrated port is immediately activated on "
                         "the destination host.")],
               help=_('Activation strategy to use for live migration.')),
]

nb_global_opts = [
    cfg.BoolOpt('ignore_lsp_down',
                default=False,
                help=_('If set to False, ARP/ND reply flows for logical '
                       'switch ports will be installed only if the port is '
                       'UP, i.e. claimed by a Chassis. If set to True, these '
                       'flows are installed regardless of the status of the '
                       'port, which can result in a situation that an ARP '
                       'request to an IP is resolved even before the relevant '
                       'VM/container is running. For environments where this '
                       'is not an issue, setting it to True can reduce '
                       'the load and latency of the control plane. '
                       'The default value is False.')),
    cfg.IntOpt('fdb_removal_limit',
               min=0,
               default=0,
               help=_('FDB aging bulk removal limit. This limits how many '
                      'rows can expire in a single transaction. Default '
                      'is 0, which is unlimited. When the limit is reached, '
                      'the next batch removal is delayed by 5 seconds. '
                      'This is supported by OVN >= 23.09.')),
    cfg.IntOpt('mac_binding_removal_limit',
               min=0,
               default=0,
               help=_('MAC binding aging bulk removal limit. This limits how '
                      'many entries can expire in a single transaction. '
                      'The default is 0 which is unlimited. When the limit '
                      'is reached, the next batch removal is delayed by '
                      '5 seconds.')),
]


def register_opts():
    cfg.CONF.register_opts(ovn_opts, group='ovn')
    ovs_conf.register_ovs_agent_opts()
    cfg.CONF.register_opts(nb_global_opts, group=OVN_NB_GLOBAL)


def list_opts():
    return [
        ('ovn', ovn_opts),
        ('ovs', ovs_conf.OPTS),
        (OVN_NB_GLOBAL, nb_global_opts),
    ]


def get_ovn_nb_connection():
    return ','.join(cfg.CONF.ovn.ovn_nb_connection)


def get_ovn_nb_private_key():
    return cfg.CONF.ovn.ovn_nb_private_key


def get_ovn_nb_certificate():
    return cfg.CONF.ovn.ovn_nb_certificate


def get_ovn_nb_ca_cert():
    return cfg.CONF.ovn.ovn_nb_ca_cert


def get_ovn_sb_connection():
    return ','.join(cfg.CONF.ovn.ovn_sb_connection)


def get_ovn_sb_private_key():
    return cfg.CONF.ovn.ovn_sb_private_key


def get_ovn_sb_certificate():
    return cfg.CONF.ovn.ovn_sb_certificate


def get_ovn_sb_ca_cert():
    return cfg.CONF.ovn.ovn_sb_ca_cert


def get_ovn_ovsdb_timeout():
    return cfg.CONF.ovn.ovsdb_connection_timeout


def get_ovn_ovsdb_retry_max_interval():
    return cfg.CONF.ovn.ovsdb_retry_max_interval


def get_ovn_ovsdb_probe_interval():
    return cfg.CONF.ovn.ovsdb_probe_interval


def get_ovn_neutron_sync_mode():
    return cfg.CONF.ovn.neutron_sync_mode


def get_ovn_l3_scheduler():
    return cfg.CONF.ovn.ovn_l3_scheduler


def is_ovn_distributed_floating_ip():
    return cfg.CONF.ovn.enable_distributed_floating_ip


def get_ovn_lm_activation_strategy():
    return cfg.CONF.ovn.live_migration_activation_strategy


def get_ovn_vhost_sock_dir():
    return cfg.CONF.ovn.vhost_sock_dir


def get_ovn_dhcp_default_lease_time():
    return cfg.CONF.ovn.dhcp_default_lease_time


def get_ovn_ovsdb_log_level():
    return VLOG_LEVELS[cfg.CONF.ovn.ovsdb_log_level]


def is_ovn_metadata_enabled():
    return cfg.CONF.ovn.ovn_metadata_enabled


def get_dns_servers():
    return cfg.CONF.ovn.dns_servers


def is_dns_records_ovn_owned():
    return cfg.CONF.ovn.dns_records_ovn_owned


def get_global_dhcpv4_opts():
    return cfg.CONF.ovn.ovn_dhcp4_global_options


def get_global_dhcpv6_opts():
    return cfg.CONF.ovn.ovn_dhcp6_global_options


def is_ovn_emit_need_to_frag_enabled():
    return cfg.CONF.ovn.ovn_emit_need_to_frag


def is_ovn_dhcp_disabled_for_baremetal():
    return cfg.CONF.ovn.disable_ovn_dhcp_for_baremetal_ports


def is_learn_fdb_enabled():
    return cfg.CONF.ovn.localnet_learn_fdb


def get_fdb_age_threshold():
    return str(cfg.CONF.ovn.fdb_age_threshold)


def get_fdb_removal_limit():
    return str(cfg.CONF.ovn_nb_global.fdb_removal_limit)


def get_ovn_mac_binding_age_threshold():
    # This value is always stored as a string in the OVN DB
    return str(cfg.CONF.ovn.mac_binding_age_threshold)


def get_ovn_mac_binding_removal_limit():
    return str(cfg.CONF.ovn_nb_global.mac_binding_removal_limit)


def is_broadcast_arps_to_all_routers_enabled():
    return cfg.CONF.ovn.broadcast_arps_to_all_routers


def is_ovn_router_indirect_snat_enabled():
    return cfg.CONF.ovn.ovn_router_indirect_snat



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools
import re
import threading

from oslo_config import cfg
from oslo_log import log
from ovsdbapp.backend.ovs_idl import vlog

from neutron.agent.linux import external_process
from neutron.agent.ovn.extensions import extension_manager
from neutron.agent.ovn.metadata import agent as metadata_agent
from neutron.agent.ovn.metadata import server_socket as metadata_server
from neutron.common.ovn import constants as ovn_const
from neutron.conf.agent.database import agents_db
from neutron.conf.agent.metadata import config as meta_conf
from neutron.conf.agent.ovn.metadata import config as ovn_meta
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as config


LOG = log.getLogger(__name__)
EXT_NAME = 'metadata'
agents_db.register_db_agents_opts()
_SYNC_STATE_LOCK = threading.RLock()
CHASSIS_METADATA_LOCK = 'chassis_metadata_lock'

SB_IDL_TABLES = ['Encap',
                 'Port_Binding',
                 'Datapath_Binding',
                 'SB_Global',
                 'Chassis',
                 'Chassis_Private',
                 ]

NS_PREFIX = ovn_const.OVN_METADATA_PREFIX
MAC_PATTERN = re.compile(r'([0-9A-F]{2}[:-]){5}([0-9A-F]{2})', re.I)
OVN_VIF_PORT_TYPES = (
    "", ovn_const.LSP_TYPE_EXTERNAL, ovn_const.LSP_TYPE_LOCALPORT)

MetadataPortInfo = collections.namedtuple('MetadataPortInfo', ['mac',
                                                               'ip_addresses',
                                                               'logical_port'])


def _sync_lock(f):
    """Decorator to block all operations for a global sync call."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        with _SYNC_STATE_LOCK:
            return f(*args, **kwargs)
    return wrapped


class MetadataExtension(extension_manager.OVNAgentExtension,
                        metadata_agent.MetadataAgent):

    def __init__(self):
        super().__init__(conf=cfg.CONF)
        vlog.use_python_logger(max_level=config.get_ovn_ovsdb_log_level())
        self._process_monitor = None
        self._proxy = None
        # We'll restart all haproxy instances upon start so that they honor
        # any potential changes in their configuration.
        self.restarted_metadata_proxy_set = set()

    @staticmethod
    def _register_config_options():
        ovn_meta.register_meta_conf_opts(meta_conf.SHARED_OPTS)
        ovn_meta.register_meta_conf_opts(
            meta_conf.UNIX_DOMAIN_METADATA_PROXY_OPTS)
        ovn_meta.register_meta_conf_opts(meta_conf.METADATA_PROXY_HANDLER_OPTS)
        ovn_meta.register_meta_conf_opts(meta_conf.METADATA_RATE_LIMITING_OPTS,
                                         group=meta_conf.RATE_LIMITING_GROUP)

    def initialize(self, *args):
        self._register_config_options()
        self._process_monitor = external_process.ProcessMonitor(
            config=self.agent_api.conf, resource_type='metadata')

    @property
    def name(self):
        return 'Metadata OVN agent extension'

    @property
    def ovs_idl_events(self):
        return []

    @property
    def nb_idl_tables(self):
        return []

    @property
    def nb_idl_events(self):
        return []

    @property
    def sb_idl_tables(self):
        return SB_IDL_TABLES

    @property
    def sb_idl_events(self):
        return [metadata_agent.PortBindingUpdatedEvent,
                metadata_agent.PortBindingDeletedEvent,
                metadata_agent.SbGlobalUpdateEvent,
                metadata_agent.ChassisPrivateCreateEvent,
                ]

    # NOTE(ralonsoh): the following properties are needed during the migration
    # to the Metadata agent to the OVN agent, while sharing the code with
    # ``metadata_agent.MetadataAgent``
    @property
    def nb_idl(self):
        return self.agent_api.nb_idl

    @nb_idl.setter
    def nb_idl(self, val):
        self.agent_api.nb_idl = val

    @property
    def sb_idl(self):
        return self.agent_api.sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self.agent_api.sb_idl = val

    @property
    def ovs_idl(self):
        return self.agent_api.ovs_idl

    @property
    def conf(self):
        return self.agent_api.conf

    @property
    def chassis(self):
        return self.agent_api.chassis

    @property
    def ovn_bridge(self):
        return self.agent_api.ovn_bridge

    @_sync_lock
    def resync(self):
        """Resync the Metadata OVN agent extension.

        Reload the configuration and sync the agent again.
        """
        self.agent_api.load_config()
        self.sync()

    def start(self):
        self._load_config()

        # Launch the server that will act as a proxy between the VM's and Nova.
        self._proxy = metadata_server.UnixDomainMetadataProxy(
            self.agent_api.conf, self.agent_api.chassis,
            sb_idl=self.agent_api.sb_idl)
        self._proxy.run()

        # Do the initial sync.
        self.sync()

        # Register the agent with its corresponding Chassis
        self.register_metadata_agent()

        # Start the metadata server.
        proxy_thread = threading.Thread(target=self._proxy.wait)
        proxy_thread.start()

        # Raise the "is_started" flag.
        self._is_started = True



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from unittest import mock
import uuid

from neutron_lib import constants as n_const
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_utils import netutils
from oslo_utils import uuidutils

from neutron.agent.linux import ip_lib
from neutron.agent.linux.ip_lib import IpAddrCommand as ip_addr
from neutron.agent.linux.ip_lib import IpLinkCommand as ip_link
from neutron.agent.linux.ip_lib import IpNetnsCommand as ip_netns
from neutron.agent.linux.ip_lib import IPWrapper as ip_wrap
from neutron.agent.linux import utils as linux_utils
from neutron.agent.ovn.metadata import agent
from neutron.agent.ovn.metadata import driver
from neutron.common.ovn import constants as ovn_const
from neutron.common import utils
from neutron.conf.agent.metadata import config as meta_conf
from neutron.conf.agent.ovn.metadata import config as ovn_meta_conf
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.tests import base


OvnPortInfo = collections.namedtuple(
    'OvnPortInfo', ['datapath', 'type', 'mac', 'external_ids', 'logical_port'])


class DatapathInfo:
    def __init__(self, uuid, external_ids):
        self.uuid = uuid
        self.external_ids = external_ids

    def __hash__(self):
        return hash(self.uuid)


def makePort(datapath=None, type='', mac=None, external_ids=None,
             logical_port=None):
    return OvnPortInfo(datapath, type, mac, external_ids, logical_port)


class ConfFixture(config_fixture.Config):
    def setUp(self):
        super().setUp()
        ovn_meta_conf.register_meta_conf_opts(meta_conf.SHARED_OPTS, self.conf)
        ovn_meta_conf.register_meta_conf_opts(
            meta_conf.UNIX_DOMAIN_METADATA_PROXY_OPTS, self.conf)
        ovn_meta_conf.register_meta_conf_opts(
            meta_conf.METADATA_PROXY_HANDLER_OPTS, self.conf)
        ovn_meta_conf.register_meta_conf_opts(
            ovn_meta_conf.OVS_OPTS, self.conf, group='ovs')
        ovn_conf.register_opts()


class TestMetadataAgent(base.BaseTestCase):
    fake_conf = cfg.CONF
    fake_conf_fixture = ConfFixture(fake_conf)

    def setUp(self):
        super().setUp()
        self.useFixture(self.fake_conf_fixture)
        self.log_p = mock.patch.object(agent, 'LOG')
        self.log = self.log_p.start()
        self.agent = agent.MetadataAgent(self.fake_conf)
        self.agent.sb_idl = mock.Mock()
        self.agent.ovs_idl = mock.Mock()
        self.agent.ovs_idl.transaction = mock.MagicMock()
        self.agent._chassis = 'chassis'
        self.agent._ovn_bridge = 'br-int'
        self.ipv6_p = mock.patch.object(netutils, 'is_ipv6_enabled')
        self.ipv6 = self.ipv6_p.start()
        self.ipv6.return_value = True

        self.ports = []
        for i in range(0, 3):
            self.ports.append(makePort(
                datapath=DatapathInfo(
                    uuid=str(uuid.uuid4()),
                    external_ids={'name': 'neutron-%d' % i})))
        self.agent.sb_idl.get_ports_on_chassis.return_value = self.ports

    def test_sync(self):

        with mock.patch.object(
                self.agent, 'provision_datapath') as pdp,\
                mock.patch.object(
                    ip_lib, 'list_network_namespaces') as lnn,\
                mock.patch.object(
                    self.agent, 'teardown_datapath') as tdp:
            lnn.return_value = [ovn_const.OVN_METADATA_PREFIX + '1',
                                ovn_const.OVN_METADATA_PREFIX + '2']

            self.agent.sync()

            pdp.assert_has_calls([mock.call(p) for p in self.ports],
                                 any_order=True)
            lnn.assert_called_once_with()
            tdp.assert_not_called()

    def test_sync_teardown_namespace(self):
        """Test that sync tears down unneeded metadata namespaces."""
        with mock.patch.object(
                self.agent, 'provision_datapath') as pdp,\
                mock.patch.object(
                    ip_lib, 'list_network_namespaces') as lnn,\
                mock.patch.object(
                    self.agent, 'teardown_datapath') as tdp:
            lnn.return_value = [ovn_const.OVN_METADATA_PREFIX + '1',
                                ovn_const.OVN_METADATA_PREFIX + '2',
                                ovn_const.OVN_METADATA_PREFIX + '3',
                                'ns1', 'ns2']

            self.agent.sync()

            pdp.assert_has_calls([mock.call(p) for p in self.ports],
                                 any_order=True)
            lnn.assert_called_once_with()
            tdp.assert_called_once_with('3')

    def test_sync_teardown_namespace_does_not_crash_on_error(self):
        """Test that sync tears down unneeded metadata namespaces.
        Even if that fails it continues to provision other datapaths
        """
        with mock.patch.object(
                self.agent, 'provision_datapath') as pdp,\
                mock.patch.object(
                    ip_lib, 'list_network_namespaces',
                    return_value=['ovnmeta-1', 'ovnmeta-2', 'ovnmeta-3',
                                  'ns1', 'ns2']) as lnn,\
                mock.patch.object(
                    self.agent, 'teardown_datapath',
                    side_effect=Exception()) as tdp:
            self.agent.sync()

            pdp.assert_has_calls([mock.call(p) for p in self.ports],
                                 any_order=True)
            lnn.assert_called_once_with()
            tdp.assert_called_once_with('3')

    def test_get_networks_port_bindings(self):
        """Test get_networks_port_bindings returns only the port binding
        objects for ports with VIF type empty ('') or 'external'.
        This test simulates that this chassis has the following ports:
            * port0: datapath 1, type ''
            * port1: datapath 1, type 'external'
            * port2: datapath 1, type 'unknown'
            * port3: datapath 2, type ''
            * port4: datapath 3, type 'external'
            * port5: datapath 4, type 'unknown'

        Only port bindings from ports 0, 1, 3, and 4 are expected.
        """

        datapath_1 = DatapathInfo(uuid='uuid1',
                                  external_ids={'name': 'neutron-1'})
        datapath_2 = DatapathInfo(uuid='uuid2',
                                  external_ids={'name': 'neutron-2'})
        datapath_3 = DatapathInfo(uuid='uuid3',
                                  external_ids={'name': 'neutron-3'})
        datapath_4 = DatapathInfo(uuid='uuid4',
                                  external_ids={'name': 'neutron-4'})

        ports = [
            makePort(datapath_1, type=''),
            makePort(datapath_1, type='external'),
            makePort(datapath_1, type='unknown'),
            makePort(datapath_2, type=''),
            makePort(datapath_3, type='external'),
            makePort(datapath_4, type='unknown')
        ]

        with mock.patch.object(self.agent.sb_idl, 'get_ports_on_chassis',
                               return_value=ports):
            self.assertEqual([ports[0], ports[1], ports[3], ports[4]],
                             self.agent.get_networks_port_bindings())

    def test_teardown_datapath(self):
        """Test teardown datapath.

        Check that the VETH pair, OVS port and namespace associated to this
        namespace are deleted and the metadata proxy is destroyed.
        """
        with mock.patch.object(ip_netns, 'exists', return_value=True),\
                mock.patch.object(
                    ip_lib, 'device_exists', return_value=True),\
                mock.patch.object(
                    ip_wrap, 'garbage_collect_namespace') as garbage_collect,\
                mock.patch.object(
                    ip_wrap, 'del_veth') as del_veth,\
                mock.patch.object(agent.MetadataAgent, '_get_veth_name',
                                  return_value=['veth_0', 'veth_1']),\
                mock.patch.object(
                    driver.MetadataDriver,
                    'destroy_monitored_metadata_proxy') as destroy_mdp:

            self.agent.teardown_datapath('1')

            destroy_mdp.assert_called_once_with(
                mock.ANY, '1', mock.ANY, ovn_const.OVN_METADATA_PREFIX + '1')
            self.agent.ovs_idl.del_port.assert_called_once_with('veth_0')
            del_veth.assert_called_once_with('veth_0')
            garbage_collect.assert_called_once_with()

    def _test__process_cidrs_when_current_namespace_empty(self, ipv6_enabled):
        """Current namespace contains no cidrs, it is missing required cidrs.
        """
        self.ipv6.return_value = ipv6_enabled
        current_namespace_cidrs = set()
        datapath_port_ips = ['10.0.0.2', '10.0.0.3', '10.0.1.5']
        metadaport_subnet_cidrs = ['10.0.0.0/30', '10.0.1.0/28', '11.0.1.2/24']
        lla = 'fe80::f816:3eff:fe63:8dc5/64'

        expected_cidrs_to_add = {'10.0.0.0/30', '10.0.1.0/28',
                                 n_const.METADATA_CIDR}
        expected_cidrs_to_delete = set()

        if ipv6_enabled:
            datapath_port_ips.extend(['2001:db8::2'])
            metadaport_subnet_cidrs.extend(['2001:db8::2/64'])
            expected_cidrs_to_add.add('2001:db8::2/64')
            expected_cidrs_to_add.add(n_const.METADATA_V6_CIDR)
            expected_cidrs_to_add.add(lla)

        actual_result = self.agent._process_cidrs(current_namespace_cidrs,
                                                  datapath_port_ips,
                                                  metadaport_subnet_cidrs,
                                                  lla)
        actual_cidrs_to_add, actual_cidrs_to_delete = actual_result

        self.assertSetEqual(actual_cidrs_to_add, expected_cidrs_to_add)
        self.assertSetEqual(actual_cidrs_to_delete, expected_cidrs_to_delete)

    def test__process_cidrs_when_current_namespace_empty_ipv6(self):
        self._test__process_cidrs_when_current_namespace_empty(True)

    def test__process_cidrs_when_current_namespace_empty_no_ipv6(self):
        self._test__process_cidrs_when_current_namespace_empty(False)

    def _test__process_cidrs_current_ns_only_contains_meta_cidr(self,
            ipv6_enabled):
        """Current namespace cidrs only contains IPv4 metadata cidr,
        and it is missing new required cidrs.
        """
        self.ipv6.return_value = ipv6_enabled
        current_namespace_cidrs = {n_const.METADATA_CIDR}
        datapath_port_ips = ['10.0.0.2', '10.0.0.3', '10.0.1.5']
        metadaport_subnet_cidrs = ['10.0.0.0/30', '10.0.1.0/28', '11.0.1.2/24']
        lla = 'fe80::f816:3eff:fe63:8dc5/64'

        expected_cidrs_to_add = {'10.0.0.0/30', '10.0.1.0/28'}
        expected_cidrs_to_delete = set()

        if ipv6_enabled:
            datapath_port_ips.extend(['2001:db8::2'])
            metadaport_subnet_cidrs.extend(['2001:db8::2/64'])
            expected_cidrs_to_add.add('2001:db8::2/64')
            expected_cidrs_to_add.add(n_const.METADATA_V6_CIDR)
            expected_cidrs_to_add.add(lla)

        actual_result = self.agent._process_cidrs(current_namespace_cidrs,
                                                  datapath_port_ips,
                                                  metadaport_subnet_cidrs,
                                                  lla)
        actual_cidrs_to_add, actual_cidrs_to_delete = actual_result

        self.assertSetEqual(actual_cidrs_to_add, expected_cidrs_to_add)
        self.assertSetEqual(actual_cidrs_to_delete, expected_cidrs_to_delete)

    def test__process_cidrs_current_ns_only_contains_meta_cidr_ipv6(self):
        self._test__process_cidrs_current_ns_only_contains_meta_cidr(True)

    def test__process_cidrs_current_ns_only_contains_meta_cidr_no_ipv6(self):
        self._test__process_cidrs_current_ns_only_contains_meta_cidr(False)

    def _test__process_cidrs_current_ns_contains_stale_cidr(self,
            ipv6_enabled):
        """Current namespace cidrs contains stale cidrs, including lla,
        and it is missing new required cidrs.
        """
        self.ipv6.return_value = ipv6_enabled
        lla = 'fe80::f816:3eff:fe63:8dc5/64'
        current_namespace_cidrs = {n_const.METADATA_CIDR, '10.0.1.0/31',
                                   lla}
        datapath_port_ips = ['10.0.0.2', '10.0.0.3', '10.0.1.5']
        metadaport_subnet_cidrs = ['10.0.0.0/30', '10.0.1.0/28', '11.0.1.2/24']

        expected_cidrs_to_add = {'10.0.0.0/30', '10.0.1.0/28'}
        expected_cidrs_to_delete = {'10.0.1.0/31'}
        # If IPv6 is enabled, the IPv6 metadata address will be added as the
        # LLA is already present in the namespace, else the LLA will be removed
        # as there are no IPv6 subnets configured.
        if ipv6_enabled:
            expected_cidrs_to_add.add(n_const.METADATA_V6_CIDR)
        else:
            expected_cidrs_to_delete.add(lla)
            # Do not pass the LLA to _process_cidrs(), indicating it should
            # not be configured
            lla = None

        actual_result = self.agent._process_cidrs(current_namespace_cidrs,
                                                  datapath_port_ips,
                                                  metadaport_subnet_cidrs,
                                                  lla)
        actual_cidrs_to_add, actual_cidrs_to_delete = actual_result

        self.assertSetEqual(actual_cidrs_to_add, expected_cidrs_to_add)
        self.assertSetEqual(actual_cidrs_to_delete, expected_cidrs_to_delete)

    def test__process_cidrs_current_ns_contains_stale_cidr_ipv6(self):
        self._test__process_cidrs_current_ns_contains_stale_cidr(True)

    def test__process_cidrs_current_ns_contains_stale_cidr_no_ipv6(self):
        self._test__process_cidrs_current_ns_contains_stale_cidr(False)

    def _test__process_cidrs_current_namespace_contains_mix_cidrs(self,
            ipv6_enabled):
        """Current namespace cidrs contains stale cidrs and it is missing
        new required cidrs.
        """
        self.ipv6.return_value = ipv6_enabled
        lla = 'fe80::f816:3eff:fe63:8dc5/64'
        current_namespace_cidrs = {n_const.METADATA_CIDR,
                                   '10.0.1.0/31',
                                   '10.0.1.0/28',
                                   'fe77::/64',
                                   '2001:db8::2/64',
                                   lla}
        datapath_port_ips = ['10.0.0.2', '10.0.1.5']
        metadaport_subnet_cidrs = ['10.0.0.0/30', '10.0.1.0/28', '11.0.1.2/24']
        expected_cidrs_to_add = {'10.0.0.0/30'}
        expected_cidrs_to_delete = {'10.0.1.0/31', 'fe77::/64'}

        if ipv6_enabled:
            datapath_port_ips.extend(['2001:db8::2'])
            metadaport_subnet_cidrs.extend(['2001:db8::2/64'])
            expected_cidrs_to_add.add(n_const.METADATA_V6_CIDR)
        else:
            expected_cidrs_to_delete.add('2001:db8::2/64')
            expected_cidrs_to_delete.add(lla)

        actual_result = self.agent._process_cidrs(current_namespace_cidrs,
                                                  datapath_port_ips,
                                                  metadaport_subnet_cidrs,
                                                  lla)
        actual_cidrs_to_add, actual_cidrs_to_delete = actual_result

        self.assertSetEqual(actual_cidrs_to_add, expected_cidrs_to_add)
        self.assertSetEqual(actual_cidrs_to_delete, expected_cidrs_to_delete)

    def test__process_cidrs_current_namespace_contains_mix_cidrs_ipv6(self):
        self._test__process_cidrs_current_namespace_contains_mix_cidrs(True)

    def test__process_cidrs_current_namespace_contains_mix_cidrs_no_ipv6(self):
        self._test__process_cidrs_current_namespace_contains_mix_cidrs(False)

    def test__get_provision_params_returns_none_when_metadata_port_is_missing(
            self):
        """Should return None when there is no metadata port in datapath and
        call teardown datapath.
        """
        network_id = '1'
        datapath = DatapathInfo(uuid='test123',
                                external_ids={'name': f'neutron-{network_id}'})

        with mock.patch.object(
                self.agent.sb_idl, 'get_metadata_port',
                return_value=None),\
            mock.patch.object(
                self.agent, 'teardown_datapath') as tdp:
            self.assertIsNone(self.agent._get_provision_params(datapath))
            tdp.assert_called_once_with(network_id)

    def test__get_provision_params_returns_none_when_metadata_port_missing_mac(
            self):
        """Should return None when metadata port is missing MAC and
        call teardown datapath.
        """
        network_id = '1'
        datapath = DatapathInfo(uuid='test123',
                                external_ids={'name': f'neutron-{network_id}'})
        metadadata_port = makePort(datapath,
                                   mac=['NO_MAC_HERE 1.2.3.4'],
                                   external_ids={'neutron:cidrs':
                                                 '10.204.0.10/29'})

        with mock.patch.object(
                self.agent.sb_idl, 'get_metadata_port',
                return_value=metadadata_port),\
            mock.patch.object(
                self.agent, 'teardown_datapath') as tdp:
            self.assertIsNone(self.agent._get_provision_params(datapath))
            tdp.assert_called_once_with(network_id)

    def test__get_provision_params_returns_none_when_no_vif_ports(self):
        """Should return None when there are no datapath ports with type
        "external" or ""(blank) and call teardown datapath.
        """
        network_id = '1'
        datapath = DatapathInfo(uuid='test123',
                                external_ids={'name': f'neutron-{network_id}'})
        datapath_ports = [makePort(datapath, type='not_vif_type')]
        metadadata_port = makePort(datapath,
                                   mac=['fa:16:3e:22:65:18 1.2.3.4'],
                                   external_ids={'neutron:cidrs':
                                                 '10.204.0.10/29'})

        with mock.patch.object(self.agent.sb_idl, 'get_metadata_port',
                               return_value=metadadata_port),\
                mock.patch.object(self.agent.sb_idl, 'get_ports_on_chassis',
                                  return_value=datapath_ports),\
                mock.patch.object(self.agent, 'teardown_datapath') as tdp:
            self.assertIsNone(self.agent._get_provision_params(datapath))
            tdp.assert_called_once_with(network_id)

    def _test__get_provision_params_returns_provision_parameters(self,
                                                                 port_ip):
        """The happy path when datapath has ports with "external" or ""(blank)
        types and metadata port contains MAC and subnet CIDRs.
        """
        network_id = '1'
        metada_port_mac = "fa:16:3e:22:65:18"
        metada_port_subnet_cidr = "10.204.0.10/29"
        metada_port_logical_port = "3b66c176-199b-48ec-8331-c1fd3f6e2b44"

        datapath = DatapathInfo(uuid='test123',
                                external_ids={'name': f'neutron-{network_id}'})
        datapath_ports = [makePort(datapath,
                                   mac=[f'fa:16:3e:e7:ac {port_ip}'])]
        metadadata_port = makePort(datapath,
                                   mac=[
                                       f'{metada_port_mac} 10.204.0.1'
                                   ],
                                   external_ids={'neutron:cidrs':
                                                 metada_port_subnet_cidr},
                                   logical_port=metada_port_logical_port)

        with mock.patch.object(self.agent.sb_idl, 'get_metadata_port',
                               return_value=metadadata_port),\
            mock.patch.object(self.agent.sb_idl, 'get_ports_on_chassis',
                              return_value=datapath_ports):
            actual_params = self.agent._get_provision_params(datapath)

        net_name, datapath_port_ips, any_ip6, metadata_port_info = (
            actual_params)

        self.assertEqual(network_id, net_name)

        if utils.get_ip_version(port_ip) == n_const.IP_VERSION_4:
            self.assertListEqual([port_ip], datapath_port_ips)
            self.assertFalse(any_ip6)
        else:
            self.assertTrue(any_ip6)
        self.assertEqual(metada_port_mac, metadata_port_info.mac)
        self.assertSetEqual({metada_port_subnet_cidr},
                            metadata_port_info.ip_addresses)
        self.assertEqual(metada_port_logical_port,
                         metadata_port_info.logical_port)

    def test__get_provision_params_returns_provision_parameters(self):
        self._test__get_provision_params_returns_provision_parameters(
            '1.2.3.4')

    def test__get_provision_params_returns_provision_parameters_ipv6(self):
        self._test__get_provision_params_returns_provision_parameters(
            'fe80::f816:3eff:feb6:c0c0')

    def _test_provision_datapath(self, ipv6_enabled):
        """Test datapath provisioning.

        Check that the VETH pair, OVS port and namespace associated to this
        namespace are created, that the interface is properly configured with
        the right IP addresses and that the metadata proxy is spawned.
        """
        self.ipv6.return_value = ipv6_enabled
        net_name = '123'
        metadaport_logical_port = '123-abc-456'
        datapath_ports_ips = ['10.0.0.1', '10.0.0.2']
        metada_port_info = agent.MetadataPortInfo(
            mac='aa:bb:cc:dd:ee:ff',
            ip_addresses=['10.0.0.1/23',
                          '2001:470:9:1224:5595:dd51:6ba2:e788/64'],
            logical_port=metadaport_logical_port
        )
        provision_params = (net_name, datapath_ports_ips, ipv6_enabled,
                            metada_port_info,)
        namespace_name = 'namespace'

        if ipv6_enabled:
            datapath_ports_ips.extend(['2001:470:9:1224:5595:dd51:6ba2:e788'])

        with mock.patch.object(self.agent,
                               '_get_provision_params',
                               return_value=provision_params),\
                mock.patch.object(
                    ip_lib, 'device_exists', return_value=False),\
                mock.patch.object(
                    ip_lib.IPDevice, 'exists', return_value=False),\
                mock.patch.object(agent.MetadataAgent, '_get_veth_name',
                                  return_value=['veth_0', 'veth_1']),\
                mock.patch.object(agent.MetadataAgent, '_get_namespace_name',
                                  return_value=namespace_name),\
                mock.patch.object(ip_link, 'set_up') as link_set_up,\
                mock.patch.object(ip_link, 'set_address') as link_set_addr, \
                mock.patch.object(ip_link, 'set_mtu') as link_set_mtu, \
                mock.patch.object(ip_addr, 'list', return_value=[]),\
                mock.patch.object(
                    ip_addr, 'add_multiple') as ip_addr_add_multiple,\
                mock.patch.object(
                    ip_wrap, 'add_veth',
                    return_value=[ip_lib.IPDevice('ip1'),
                                  ip_lib.IPDevice('ip2')]) as add_veth,\
                mock.patch.object(linux_utils, 'delete_if_exists'), \
                mock.patch.object(
                    driver.MetadataDriver,
                    'spawn_monitored_metadata_proxy') as spawn_mdp, \
                mock.patch.object(
                    self.agent, '_ensure_datapath_checksum') as mock_checksum:

            # Simulate that the VETH pair was already present in 'br-fake'.
            # We need to assert that it was deleted first.
            self.agent.ovs_idl.list_br.return_value.execute.return_value = (
                ['br-int', 'br-fake'])
            mtu = 1500
            port_binding = mock.Mock(
                datapath='fake_datapath',
                external_ids={ovn_const.OVN_NETWORK_MTU_EXT_ID_KEY: str(mtu)})
            self.agent.provision_datapath(port_binding)

            # Check that the port was deleted from br-fake
            self.agent.ovs_idl.del_port.assert_called_once_with(
                'veth_0', bridge='br-fake', if_exists=True)
            # Check that the VETH pair is created
            add_veth.assert_called_once_with('veth_0', 'veth_1',
                                             namespace_name)
            # Make sure that the two ends of the VETH pair have been set as up.
            self.assertEqual(2, link_set_up.call_count)
            link_set_mtu.assert_has_calls([mock.call(mtu), mock.call(mtu)])
            link_set_addr.assert_called_once_with('aa:bb:cc:dd:ee:ff')
            # Make sure that the port has been added to OVS.
            self.agent.ovs_idl.add_port.assert_called_once_with(
                'br-int', 'veth_0')
            self.agent.ovs_idl.db_set.assert_called_once_with(
                'Interface', 'veth_0',
                ('external_ids', {'iface-id': metadaport_logical_port}))
            # Check that the metadata port has the IP addresses properly
            # configured and that IPv6 address has been skipped.
            expected_call = [n_const.METADATA_CIDR, '10.0.0.1/23']
            bind_address_v6 = None
            if ipv6_enabled:
                expected_call.extend(
                    ['2001:470:9:1224:5595:dd51:6ba2:e788/64',
                     n_const.METADATA_V6_CIDR,
                     ip_lib.get_ipv6_lladdr('aa:bb:cc:dd:ee:ff')])
                bind_address_v6 = n_const.METADATA_V6_IP
            self.assertCountEqual(expected_call,
                                  ip_addr_add_multiple.call_args.args[0])
            # Check that metadata proxy has been spawned
            spawn_mdp.assert_called_once_with(
                mock.ANY, namespace_name, 80, mock.ANY,
                bind_address=n_const.METADATA_V4_IP, network_id=net_name,
                bind_address_v6=bind_address_v6,
                bind_interface='veth_1')
            mock_checksum.assert_called_once_with(namespace_name)

    def test_provision_datapath_ipv6(self):
        self._test_provision_datapath(True)

    def test_provision_datapath_no_ipv6(self):
        self._test_provision_datapath(False)

    def test__load_config(self):
        # Chassis name UUID formatted string. OVN bridge "br-ovn".
        valid_uuid_str = uuidutils.generate_uuid()
        self.agent.ovs_idl.db_get.return_value.execute.side_effect = [
            {'system-id': valid_uuid_str}, {'ovn-bridge': 'br-ovn'}]
        self.agent._load_config()
        self.assertEqual(valid_uuid_str, self.agent.chassis)
        self.assertEqual(uuid.UUID(valid_uuid_str), self.agent.chassis_id)
        self.assertEqual('br-ovn', self.agent.ovn_bridge)

        # Chassis name non UUID formatted string. OVN bridge not defined,
        # "br-int" assigned by default.
        self.agent.ovs_idl.db_get.return_value.execute.side_effect = [
            {'system-id': 'RandomName1'}, {}]
        self.agent._load_config()
        generated_uuid = uuid.uuid5(agent.OVN_METADATA_UUID_NAMESPACE,
                                    'RandomName1')
        self.assertEqual('RandomName1', self.agent.chassis)
        self.assertEqual(generated_uuid, self.agent.chassis_id)
        self.assertEqual('br-int', self.agent.ovn_bridge)



# Copyright 2025 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import socketserver
import urllib

import jinja2
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
import requests
import webob
from webob import exc as webob_exc

from neutron._i18n import _
from neutron.agent.linux import utils as agent_utils
from neutron.agent.metadata import proxy_base
from neutron.common import ipv6_utils
from neutron.common.ovn import constants as ovn_const
from neutron.common import utils as common_utils


LOG = logging.getLogger(__name__)

RESPONSE = jinja2.Template("""HTTP/1.1 {{ http_code }}
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: {{ len }}

<html>
 <head>
  <title>{{ title }}</title>
 </head>
 <body>
  <h1>{{ body_title }}</h1>
  {{ body }}<br /><br />


 </body>
</html>""")
RESPONSE_LENGHT = 40


class MetadataProxyHandlerBaseSocketServer(
        proxy_base.MetadataProxyHandlerBase):
    @staticmethod
    def _http_response(http_response, request):
        _res = webob.Response(
            body=http_response.content,
            status=http_response.status_code,
            content_type=http_response.headers['content-type'],
            charset=http_response.encoding)
        # NOTE(ralonsoh): there should be a better way to format the HTTP
        # response, adding the HTTP version to the ``webob.Response``
        # output string.
        out = request.http_version + ' ' + str(_res)
        if (int(_res.headers['content-length']) == 0 and
                _res.status_code == 200):
            # Add 2 extra \r\n to the result. HAProxy is also expecting
            # it even when the body is empty.
            out += '\r\n\r\n'
        return out.encode(http_response.encoding)

    def _proxy_request(self, instance_id, project_id, req):
        headers = {
            'X-Forwarded-For': req.headers.get('X-Forwarded-For'),
            'X-Instance-ID': instance_id,
            'X-Tenant-ID': project_id,
            'X-Instance-ID-Signature': common_utils.sign_instance_id(
                self.conf, instance_id)
        }

        nova_host_port = ipv6_utils.valid_ipv6_url(
            self.conf.nova_metadata_host,
            self.conf.nova_metadata_port)

        url = urllib.parse.urlunsplit((
            self.conf.nova_metadata_protocol,
            nova_host_port,
            req.path_info,
            req.query_string,
            ''))

        disable_ssl_certificate_validation = self.conf.nova_metadata_insecure
        if self.conf.auth_ca_cert and not disable_ssl_certificate_validation:
            verify_cert = self.conf.auth_ca_cert
        else:
            verify_cert = not disable_ssl_certificate_validation

        client_cert = None
        if self.conf.nova_client_cert and self.conf.nova_client_priv_key:
            client_cert = (self.conf.nova_client_cert,
                           self.conf.nova_client_priv_key)

        try:
            resp = requests.request(method=req.method, url=url,
                                    headers=headers,
                                    data=req.body,
                                    cert=client_cert,
                                    verify=verify_cert,
                                    timeout=60)
        except requests.ConnectionError:
            msg = _('The remote metadata server is temporarily unavailable. '
                    'Please try again later.')
            LOG.warning(msg)
            title = '503 Service Unavailable'
            length = RESPONSE_LENGHT + len(title) * 2 + len(msg)
            reponse = RESPONSE.render(http_code=title, title=title,
                                      body_title=title, body=title, len=length)
            return encodeutils.to_utf8(reponse)

        if resp.status_code == 200:
            return self._http_response(resp, req)
        if resp.status_code == 403:
            LOG.warning(
                'The remote metadata server responded with Forbidden. This '
                'response usually occurs when shared secrets do not match.'
            )
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code == 500:
            msg = _(
                'Remote metadata server experienced an internal server error.'
            )
            LOG.warning(msg)
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code in (400, 404, 409, 502, 503, 504):
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        raise Exception(_('Unexpected response code: %s') % resp.status_code)


class MetadataProxyHandler(MetadataProxyHandlerBaseSocketServer,
                           socketserver.StreamRequestHandler):
    NETWORK_ID_HEADER = 'X-OVN-Network-ID'
    ROUTER_ID_HEADER = ''
    _conf = None
    _chassis = None
    _sb_idl = None

    def __init__(self, request, client_address, server):
        super().__init__(self._conf, has_cache=False, request=request,
                         client_address=client_address, server=server)

    def handle(self):
        try:
            request = self.request.recv(4096)
            LOG.debug('Request: %s', request.decode('utf-8'))
            f_request = io.BytesIO(request)
            req = webob.Request.from_file(f_request)
            instance_id, project_id = self._get_instance_and_project_id(req)
            if instance_id:
                res = self._proxy_request(instance_id, project_id, req)
                self.wfile.write(res)
                return
            # TODO(ralonsoh): change this return to be a formatted Request
            # and added to self.wfile
            return webob_exc.HTTPNotFound()
        except Exception as exc:
            LOG.exception('Error while receiving data.')
            raise exc

    @property
    def sb_idl(self):
        return self._sb_idl

    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        ports = self.sb_idl.get_network_port_bindings_by_ip(network_id,
                                                            remote_address,
                                                            mac=remote_mac)
        num_ports = len(ports)
        if num_ports == 1:
            external_ids = ports[0].external_ids
            return (external_ids[ovn_const.OVN_DEVID_EXT_ID_KEY],
                    external_ids[ovn_const.OVN_PROJID_EXT_ID_KEY])
        if num_ports == 0:
            LOG.error("No port found in network %s with IP address %s",
                      network_id, remote_address)
        elif num_ports > 1:
            port_uuids = ', '.join([str(port.uuid) for port in ports])
            LOG.error("More than one port found in network %s with IP address "
                      "%s. Please run the neutron-ovn-db-sync-util script as "
                      "there seems to be inconsistent data between Neutron "
                      "and OVN databases. OVN Port uuids: %s", network_id,
                      remote_address, port_uuids)
        return None, None


class UnixDomainMetadataProxy(proxy_base.UnixDomainMetadataProxyBase):
    def __init__(self, conf, chassis, sb_idl=None):
        super().__init__(conf)
        self.chassis = chassis
        self.sb_idl = sb_idl
        agent_utils.ensure_directory_exists_without_file(
            cfg.CONF.metadata_proxy_socket)
        self._server = None

    def run(self):
        file_socket = cfg.CONF.metadata_proxy_socket
        self._server = socketserver.ThreadingUnixStreamServer(
            file_socket, MetadataProxyHandler)
        MetadataProxyHandler._conf = self.conf
        MetadataProxyHandler._chassis = self.chassis
        MetadataProxyHandler._sb_idl = self.sb_idl

    def wait(self):
        self._server.serve_forever()



# Copyright (c) 2023 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import port_trusted_vif as apidef
from neutron_lib.api.definitions import portbindings
from neutron_lib import context
from oslo_config import cfg

from neutron.db import port_trusted_db
from neutron.plugins.ml2 import plugin
from neutron.tests.common import test_db_base_plugin_v2


class PortTrustedExtensionTestPlugin(
        plugin.Ml2Plugin,
        port_trusted_db.PortTrustedDbMixin):
    """Test plugin to mixin the port trusted extension."""

    supported_extension_aliases = [apidef.ALIAS, portbindings.ALIAS]


@ddt.ddt
class PortTrustedExtensionTestCase(
         test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    """Test API extension port-trusted-vif attributes."""

    def setUp(self, *args):
        plugin = ('neutron.tests.unit.extensions.test_port_trusted_vif.'
                  'PortTrustedExtensionTestPlugin')
        extension_drivers = ['port_trusted']
        cfg.CONF.set_override('extension_drivers', extension_drivers, 'ml2')
        super().setUp(plugin=plugin)
        self.ctx = context.get_admin_context()

    def _create_and_check_port_with_trusted_field(self, trusted):
        name = 'port-trusted-vif'
        keys = [('name', name),
                ('admin_state_up', True),
                ('trusted', trusted)]
        port_args = {'name': name}
        if trusted is not None:
            port_args['trusted'] = trusted
        with self.port(is_admin=True, **port_args) as port:
            for k, v in keys:
                self.assertEqual(v, port['port'][k])
            if trusted is not None:
                self.assertEqual(trusted,
                                 port['port']['binding:profile']['trusted'])
            else:
                self.assertNotIn('trusted',
                                 port['port']['binding:profile'].keys())

    def test_create_port_with_trusted_field(self):
        self._create_and_check_port_with_trusted_field(True)
        self._create_and_check_port_with_trusted_field(False)

    def test_create_port_with_trusted_field_not_set(self):
        self._create_and_check_port_with_trusted_field(None)



# Copyright (c) 2023 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import context
from neutron_lib.plugins.ml2 import ovs_constants as p_const
from oslo_config import cfg

from neutron.agent.common import ovs_lib
from neutron.agent.l2.extensions.metadata import metadata_flows_process
from neutron.agent.l2.extensions.metadata import metadata_path
from neutron.api.rpc.callbacks import resources
from neutron.conf.plugins.ml2.drivers import ovs_conf
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_api
from neutron.tests import base


class MetadataPathAgentExtensionTestCase(base.BaseTestCase):

    def setUp(self):
        super(MetadataPathAgentExtensionTestCase, self).setUp()
        ovs_conf.register_ovs_agent_opts(cfg=cfg.CONF)
        cfg.CONF.set_override('provider_cidr', '240.0.0.0/31', 'METADATA')
        self.context = context.get_admin_context()
        self.int_br = mock.Mock()
        self.meta_br = mock.Mock()
        self.plugin_rpc = mock.Mock()
        self.remote_resource_cache = mock.Mock()
        self.plugin_rpc.remote_resource_cache = self.remote_resource_cache
        self.meta_ext = metadata_path.MetadataPathAgentExtension()
        self.bridge_mappings = {"meta": "br-meta"}
        self.int_ofport = 200
        self.phys_ofport = 100
        self.tap_meta_ofport = 300
        self.agent_api = ovs_ext_api.OVSAgentExtensionAPI(
            self.int_br,
            tun_br=mock.Mock(),
            phys_brs={"meta": self.meta_br},
            plugin_rpc=self.plugin_rpc,
            phys_ofports={"meta": self.phys_ofport},
            bridge_mappings=self.bridge_mappings)
        self.meta_ext.consume_api(self.agent_api)
        self.m_install_arp_responder = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension.install_arp_responder").start()
        mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension.metadata_path_defaults").start()
        mock.patch(
            "neutron.agent.linux.ip_lib.IpLinkCommand.set_address").start()
        mock.patch(
            "neutron.agent.linux.ip_lib.IpAddrCommand.add").start()
        mock.patch(
            "neutron.agent.linux.ip_lib.IpLinkCommand.set_up").start()
        self.meta_ext._set_port_vlan = mock.Mock()
        self.meta_ext.initialize(None, None)
        # set int_br back to mock
        self.meta_ext.int_br = self.int_br
        # set meta_br back to mock
        self.meta_ext.meta_br = self.meta_br
        self.meta_ext.ofport_int_to_meta = self.int_ofport
        self.meta_ext.metadata_ofport = self.tap_meta_ofport
        mock.patch.object(
            self.int_br, 'get_port_ofport',
            return_value=self.int_ofport).start()
        mock.patch.object(
            self.meta_br, 'get_port_ofport',
            return_value=self.phys_ofport).start()

        self.meta_daemon = mock.Mock()
        self.meta_ext.meta_daemon = mock.Mock()

        self.port_provider_ip = "100.100.100.100"
        self.port_provider_mac = "fa:16:ee:11:22:33"
        self.port_local_vlan = 1

        def m_get_value_from_ovsdb_other_config(p, key, value_type=None):
            if key == "provider_ip":
                return self.port_provider_ip
            if key == "provider_mac":
                return self.port_provider_mac
            if key == "tag":
                return self.port_local_vlan

        self.meta_ext.rcache_api = mock.Mock()

        mock.patch.object(
            self.int_br, 'get_value_from_other_config',
            side_effect=m_get_value_from_ovsdb_other_config).start()
        mock.patch.object(
            self.int_br, 'set_value_to_other_config').start()

        mock.patch.object(
            self.meta_br, 'set_value_to_other_config').start()

        mock.patch.object(
            self.meta_ext.rcache_api, 'get_resources',
            return_value=[]).start()

        self.m_add_f_nat = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension.add_flow_snat_br_meta").start()

        self.m_in_direct_to_int = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension."
            "add_flow_ingress_dnat_direct_to_int_br").start()

        self.m_out_egress_direct = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension."
            "add_flow_int_br_egress_direct").start()

        self.m_ingress_output = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension."
            "add_flow_int_br_ingress_output").start()

        self.m_remove_direct = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension."
            "remove_port_metadata_direct_flow").start()

        self.m_remove_port_flows = mock.patch(
            "neutron.agent.l2.extensions.metadata.metadata_path."
            "MetadataPathAgentExtension."
            "remove_port_metadata_path_nat_and_arp_flow").start()

    def test_handle_port(self):
        port_mac_address = "aa:aa:aa:aa:aa:aa"
        port_name = "tap-p1"
        port_id = "p1"
        port_ofport = 1
        port_device_owner = "compute:test"
        port_ip = "1.1.1.1"
        with mock.patch.object(self.meta_ext.meta_daemon,
                 "config") as h_config, mock.patch.object(
                     self.meta_ext.ext_api,
                     "get_provider_ip_info") as get_p_info:
            get_p_info.return_value = {
                'instance_id': 'instance_uuid_1',
                'project_id': 'project_id_1',
                'provider_ip': self.port_provider_ip,
                'provider_port_mac': self.port_provider_mac
            }

            port = {"port_id": port_id,
                    "fixed_ips": [{"ip_address": port_ip,
                                   "subnet_id": "1"}],
                    "vif_port": ovs_lib.VifPort(port_name, port_ofport,
                                                port_id,
                                                port_mac_address, "br-int"),
                    "device_owner": port_device_owner,
                    "network_id": "net_id_1",
                    "mac_address": port_mac_address}
            self.meta_ext.handle_port(self.context, port)

            get_p_info.assert_called_once_with(
                port['port_id'],
                self.port_provider_ip,
                self.port_provider_mac)
            h_config.assert_called_once_with(
                list(self.meta_ext.instance_infos.values()))

        self.m_install_arp_responder.assert_has_calls(
            [mock.call(
                 bridge=mock.ANY,
                 ip=metadata_flows_process.METADATA_V4_IP,
                 mac=metadata_path.METADATA_DEFAULT_MAC,
                 table=p_const.TRANSIENT_TABLE),
             mock.call(
                 ip=self.port_provider_ip,
                 mac=self.port_provider_mac)])
        self.m_add_f_nat.assert_called_once_with(
            self.port_local_vlan, port_mac_address, port_ip,
            self.port_provider_mac, self.port_provider_ip)

        self.m_in_direct_to_int.assert_called_once_with(
            self.port_local_vlan, self.meta_ext.provider_vlan_id,
            self.port_provider_ip,
            port_mac_address,
            port_ip,
            self.phys_ofport,
            self.tap_meta_ofport)
        self.m_out_egress_direct.assert_called_once_with(
            port_ofport, self.port_local_vlan, self.int_ofport)
        self.m_ingress_output.assert_called_once_with(
            self.int_ofport,
            self.port_local_vlan,
            port_mac_address,
            port_ofport)

    def test_get_port_no_more_provider_ip(self):

        def m_get_value_from_ovsdb_other_config(p, key, value_type=None):
            if key == "provider_ip":
                return
            if key == "provider_mac":
                return

        mock.patch.object(
            self.int_br, 'get_value_from_other_config',
            side_effect=m_get_value_from_ovsdb_other_config).start()
        mock.patch.object(
            self.int_br, 'set_value_to_other_config').start()

        port_device_owner = "compute:test"

        class Port(object):
            def __init__(self):
                self.device_id = "d1"
                self.project_id = "p1"

        with mock.patch.object(self.meta_ext.meta_daemon,
                 "config"), mock.patch.object(
                     self.meta_ext.ext_api.cache_api,
                     "get_resource_by_id",
                     return_value=Port()) as get_res:
            port1_mac_address = "aa:aa:aa:aa:aa:aa"
            port1_name = "tap-p1"
            port1_id = "p1"
            port1_ofport = 1
            port1 = {"port_id": port1_id,
                     "fixed_ips": [{"ip_address": "1.1.1.1",
                                    "subnet_id": "1"}],
                     "vif_port": ovs_lib.VifPort(port1_name, port1_ofport,
                                                 port1_id,
                                                 port1_mac_address, "br-int"),
                     "device_owner": port_device_owner,
                     "network_id": "net_id_1",
                     "mac_address": port1_mac_address}
            self.meta_ext.handle_port(self.context, port1)

            get_res.assert_called_once_with(
                resources.PORT,
                port1['port_id'])

            port2_id = "p2"
            self.assertRaises(
                metadata_path.NoMoreProviderRes,
                self.meta_ext.ext_api.get_provider_ip_info,
                port2_id, None, None)

    def test_delete_port(self):
        port_mac_address = "aa:aa:aa:aa:aa:aa"
        port_name = "tap-p1"
        port_id = "p1"
        port_ofport = 1
        port_device_owner = "compute:test"
        port_ip = "1.1.1.1"
        with mock.patch.object(self.meta_ext.meta_daemon,
                 "config") as h_config:
            port = {"port_id": port_id,
                    "fixed_ips": [{"ip_address": port_ip,
                                   "subnet_id": "1"}],
                    "vif_port": ovs_lib.VifPort(port_name, port_ofport,
                                                port_id,
                                                port_mac_address, "br-int"),
                    "device_owner": port_device_owner,
                    "network_id": "net_id_1",
                    "mac_address": port_mac_address}
            self.meta_ext.handle_port(self.context, port)
            instance_info_values = list(self.meta_ext.instance_infos.values())

            self.meta_ext.delete_port(self.context, {"port_id": port_id})
            h_config.assert_has_calls([mock.call(instance_info_values),
                                       mock.call([])])
            self.assertNotIn(self.port_provider_ip,
                             self.meta_ext.ext_api.allocated_ips)
            self.assertNotIn(self.port_provider_mac,
                             self.meta_ext.ext_api.allocated_macs)

        self.m_remove_direct.assert_called_once_with(
            port_ofport, self.port_local_vlan,
            port_mac_address, self.int_ofport)
        self.m_remove_port_flows.assert_called_once_with(
            self.port_local_vlan, port_mac_address,
            port_ip, port_ofport, mock.ANY)



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno

from oslo_config import cfg

from neutron._i18n import _
from neutron.agent.linux import conntrackd
from neutron.agent.linux import external_process
from neutron.agent.linux import ip_lib
from neutron.common import utils as common_utils
from neutron.conf.agent.l3 import config as l3_config
from neutron.conf.agent.l3 import ha as ha_config
from neutron.tests.common import net_helpers
from neutron.tests.functional import base
from neutron.tests.unit.agent.l3.test_agent import FAKE_ID
from neutron.tests.unit.agent.linux.test_conntrackd import \
    ConntrackdConfigTestCase
from neutron_lib.exceptions import ProcessExecutionError


class ConntrackdManagerTestCase(base.BaseSudoTestCase):

    def setUp(self):
        super().setUp()
        l3_config.register_l3_agent_config_opts(l3_config.OPTS, cfg.CONF)
        ha_config.register_l3_agent_ha_opts()
        self.config(check_child_processes_interval=1, group='AGENT')

        self.process_monitor = external_process.ProcessMonitor(cfg.CONF,
                                                               'router')
        self.namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        self.ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
        self._prepare_device()

        self.manager = conntrackd.ConntrackdManager(
            FAKE_ID,
            self.process_monitor,
            cfg.CONF,
            '192.168.0.5',
            3,
            'eth0',
            namespace=self.namespace)
        self.addCleanup(self._stop_conntrackd_manager)

    def _stop_conntrackd_manager(self):
        try:
            self.manager.disable()
        except ProcessExecutionError as process_err:
            # self.manager.disable() will perform SIGTERM->wait->SIGKILL
            # (if needed) on the process. However, it is sometimes possible
            # that SIGKILL gets called on a process that just exited due to
            # SIGTERM. Ignore this condition so the test is not marked as
            # failed.
            if not (len(process_err.args) > 0 and
                    "No such process" in process_err.args[0]):
                raise

    def _prepare_device(self):
        # NOTE(gaudenz): this is the device used in the conntrackd config
        # file
        ip_device = self.ip_wrapper.add_dummy('eth0')
        ip_device.link.set_up()
        ip_device.addr.add('192.168.0.5/24')

    def _spawn_conntrackd(self, conntrackd_manager):
        conntrackd_manager.spawn()
        process = conntrackd_manager.get_process()
        common_utils.wait_until_true(
            lambda: process.active,
            timeout=5,
            sleep=0.01,
            exception=RuntimeError(_("Conntrackd didn't spawn")))
        return process

    def _get_conf_on_disk(self):
        config_path = self.manager.get_conffile_path()
        try:
            with open(config_path) as conf:
                return conf.read()
        except (OSError, IOError) as e:
            if e.errno != errno.ENOENT:
                raise
        return ''

    def test_conntrackd_config(self):
        self._spawn_conntrackd(self.manager)

        expected_config = ConntrackdConfigTestCase.get_expected(
            cfg.CONF.ha_confs_path,
        )
        self.assertEqual(expected_config,
                         self._get_conf_on_disk())

    def test_conntrackd_spawn(self):
        process = self._spawn_conntrackd(self.manager)

        self.assertTrue(process.active)

    def _test_conntrackd_respawns(self, normal_exit=True):
        process = self._spawn_conntrackd(self.manager)
        pid = process.pid
        exit_code = '-15' if normal_exit else '-9'

        # Exit the process, and see that when it comes back
        # It's indeed a different process
        self.ip_wrapper.netns.execute(['kill', exit_code, pid],
                                      privsep_exec=True)
        common_utils.wait_until_true(
            lambda: process.active and pid != process.pid,
            timeout=5,
            sleep=0.01,
            exception=RuntimeError(_("Conntrackd didn't respawn")))

    def test_conntrackd_respawns(self):
        self._test_conntrackd_respawns()

    def test_conntrackd_respawn_with_unexpected_exit(self):
        self._test_conntrackd_respawns(False)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import random
from unittest import mock

from neutron.tests import base

from neutron.common.ovn import constants as ovn_const
from neutron.scheduler import l3_ovn_scheduler


class FakeOVNGatewaySchedulerNbOvnIdl:
    def __init__(self, chassis_gateway_mapping, gateway):
        self.get_all_chassis_gateway_bindings = mock.Mock(
            return_value=chassis_gateway_mapping['Chassis_Bindings'])
        self.get_gateway_chassis_binding = mock.Mock(
            return_value=chassis_gateway_mapping['Gateways'].get(gateway,
                                                                 None))
        self.get_lrouter_by_lrouter_port = mock.Mock(
            return_value=None)


class FakeOVNGatewaySchedulerSbOvnIdl:
    def __init__(self, chassis_and_azs):
        self.get_chassis_and_azs = mock.Mock(return_value=chassis_and_azs)


class TestOVNGatewayScheduler(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.mock_log = mock.patch.object(l3_ovn_scheduler, 'LOG').start()

        # Overwritten by derived classes
        self.l3_scheduler = None

        # Used for unit tests
        self.new_gateway_name = 'lrp_new'
        self.fake_chassis_gateway_mappings = {
            'None': {'Chassis': [],
                     'Gateways': {'g1': None}},
            'Multiple1': {'Chassis': ['hv1', 'hv2', 'hv3', 'hv4', 'hv5'],
                          'Gateways': {
                              'g1': ['hv1', 'hv2', 'hv4', 'hv3', 'hv5'],
                              'g2': ['hv2', 'hv3', 'hv5', 'hv1', 'hv4'],
                              'g3': ['hv3', 'hv5', 'hv1', 'hv4', 'hv2'],
                              'g4': ['hv4', 'hv1', 'hv2', 'hv5', 'hv3']}},
            'Multiple2': {'Chassis': ['hv1', 'hv2', 'hv3'],
                          'Gateways': {'g1': ['hv1', 'hv2', 'hv3'],
                                       'g2': ['hv2', 'hv1', 'hv3'],
                                       'g3': ['hv2', 'hv1', 'hv3']}},
            'Multiple3': {'Chassis': ['hv1', 'hv2', 'hv3'],
                          'Gateways': {'g1': ['hv3', 'hv2', 'hv1'],
                                       'g2': ['hv2', 'hv1', 'hv3'],
                                       'g3': ['hv2', 'hv1', 'hv3']}},
            'Multiple4': {'Chassis': ['hv1', 'hv2'],
                          'Gateways': {'g1': ['hv1', 'hv2'],
                                       'g2': ['hv1'],
                                       'g3': ['hv1'],
                                       'g4': ['hv1'],
                                       'g5': ['hv1'],
                                       'g6': ['hv1']}},
            'Multiple5': {'Chassis': ['hv1', 'hv2', 'hv3', 'hv4', 'hv5'],
                          'Gateways': {
                              'g1': ['hv1', 'hv2', 'hv3', 'hv4', 'hv5'],
                              'g2': ['hv3', 'hv2', 'hv4', 'hv5', 'hv1'],
                              'g3': ['hv4', 'hv5', 'hv1', 'hv2', 'hv3'],
                              'g4': ['hv5', 'hv1', 'hv2', 'hv3', 'hv4']}},
            'Multiple6': {'Chassis': ['hv1', 'hv2', 'hv3'],
                          'Gateways': {
                              'g1': ['hv1', 'hv2', 'hv3'],
                              'g2': ['hv1', 'hv2', 'hv3'],
                              'g3': ['hv3', 'hv2', 'hv1'],
                              'g4': ['hv3', 'hv2', 'hv1']}}}

        self.fake_chassis_and_azs = {
            'None': {},
            'Multiple1': {},
            'Multiple2': {},
            'Multiple3': {},
            'Multiple4': {},
            'Multiple5': {},
            'Multiple6': {}}

        # Determine the chassis to gateway list bindings
        for details in self.fake_chassis_gateway_mappings.values():
            self.assertNotIn(self.new_gateway_name, details['Gateways'])
            details.setdefault('Chassis_Bindings', {})
            for chassis in details['Chassis']:
                details['Chassis_Bindings'].setdefault(chassis, [])
            for gw, chassis_list in details['Gateways'].items():
                chassis_list = chassis_list or []
                max_prio = len(chassis_list)
                for idx, chassis in enumerate(chassis_list):
                    prio = max_prio - idx
                    if chassis in details['Chassis_Bindings']:
                        details['Chassis_Bindings'][chassis].append((gw, prio))

    def select(self, chassis_gateway_mapping, gateway_name,
               chassis_and_azs, candidates=None):
        nb_idl = FakeOVNGatewaySchedulerNbOvnIdl(chassis_gateway_mapping,
                                                 gateway_name)
        sb_idl = FakeOVNGatewaySchedulerSbOvnIdl(chassis_and_azs)
        return self.l3_scheduler.select(nb_idl, sb_idl, gateway_name,
                                        candidates=candidates)

    def filter_existing_chassis(self, *args, **kwargs):
        return self.l3_scheduler.filter_existing_chassis(
            nb_idl=kwargs.pop('nb_idl'), gw_chassis=kwargs.pop('gw_chassis'),
            physnet=kwargs.pop('physnet'),
            chassis_physnets=kwargs.pop('chassis_physnets'),
            existing_chassis=kwargs.pop('existing_chassis'),
            az_hints=kwargs.pop('az_hints', []),
            chassis_with_azs=kwargs.pop('chassis_with_azs', {}))


class OVNGatewayChanceScheduler(TestOVNGatewayScheduler):

    def setUp(self):
        super().setUp()
        self.l3_scheduler = l3_ovn_scheduler.OVNGatewayChanceScheduler()

    def test_no_chassis_available_for_existing_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['None']
        chassis_and_azs = self.fake_chassis_and_azs['None']
        gateway_name = random.choice(list(mapping['Gateways'].keys()))
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertIsNone(chassis)

    def test_no_chassis_available_for_new_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['None']
        chassis_and_azs = self.fake_chassis_and_azs['None']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertIsNone(chassis)

    def test_random_chassis_available_for_new_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple1']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple1']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertCountEqual(chassis, mapping.get('Chassis'))

    def test_no_candidates_provided(self):
        mapping = self.fake_chassis_gateway_mappings['None']
        chassis_and_azs = self.fake_chassis_and_azs['None']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs)
        self.assertIsNone(chassis)
        self.mock_log.warning.assert_called_once_with(
            'Gateway %s was not scheduled on any chassis, no candidates are '
            'available', gateway_name)

    def test_filter_existing_chassis(self):
        # filter_existing_chassis is scheduler independent, but calling
        # it from Base class didnt seem right. Also, there is no need to have
        # another test in LeastLoadedScheduler.
        chassis_physnets = {'temp': ['phys-network-0', 'phys-network-1']}
        nb_idl = FakeOVNGatewaySchedulerNbOvnIdl(
            self.fake_chassis_gateway_mappings['None'], 'g1')
        # Check if invalid chassis is removed
        self.assertEqual(
            ['temp'], self.filter_existing_chassis(
                nb_idl=nb_idl, gw_chassis=["temp"],
                physnet='phys-network-1',
                chassis_physnets=chassis_physnets,
                existing_chassis=['temp', None]))
        # Check if invalid is removed -II
        self.assertFalse(
            self.filter_existing_chassis(
                nb_idl=nb_idl, gw_chassis=["temp"],
                physnet='phys-network-1',
                chassis_physnets=chassis_physnets,
                existing_chassis=None))
        # Check if chassis removed when physnet doesnt exist
        self.assertFalse(
            self.filter_existing_chassis(
                nb_idl=nb_idl, gw_chassis=["temp"],
                physnet='phys-network-2',
                chassis_physnets=chassis_physnets,
                existing_chassis=['temp']))
        # Check if chassis removed when it doesnt exist in gw_chassis
        # or in chassis_physnets
        self.assertFalse(
            self.filter_existing_chassis(
                nb_idl=nb_idl, gw_chassis=["temp1"],
                physnet='phys-network-2',
                chassis_physnets=chassis_physnets,
                existing_chassis=['temp']))


class OVNGatewayChanceSchedulerWithAZ(OVNGatewayChanceScheduler):

    def setUp(self):
        super().setUp()

        self.fake_chassis_and_azs = {
            'None': {},
            'Multiple1': {'hv1': {'az-1'},
                          'hv2': {'az-0'},
                          'hv3': {'az-0'},
                          'hv4': {'az-0'},
                          'hv5': {'az-0'}},
            'Multiple2': {},
            'Multiple3': {},
            'Multiple4': {},
            'Multiple5': {},
            'Multiple6': {}}


class OVNGatewayLeastLoadedScheduler(TestOVNGatewayScheduler):

    def setUp(self):
        super().setUp()
        self.l3_scheduler = l3_ovn_scheduler.OVNGatewayLeastLoadedScheduler()

    def test_no_chassis_available_for_existing_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['None']
        chassis_and_azs = self.fake_chassis_and_azs['None']
        gateway_name = random.choice(list(mapping['Gateways'].keys()))
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertIsNone(chassis)

    def test_no_chassis_available_for_new_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['None']
        chassis_and_azs = self.fake_chassis_and_azs['None']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertIsNone(chassis)

    def test_least_loaded_chassis_available_for_new_gateway1(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple1']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple1']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertCountEqual(chassis, mapping.get('Chassis'))
        # least loaded will be the first one in the list,
        # networking-ovn will assign highest priority to this first element
        self.assertEqual(['hv5', 'hv4', 'hv3', 'hv2', 'hv1'], chassis)

    def test_least_loaded_chassis_available_for_new_gateway2(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple2']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple2']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        # hv1 will have least priority
        self.assertEqual(chassis[2], 'hv1')

    def test_least_loaded_chassis_available_for_new_gateway3(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple3']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple3']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        # least loaded chassis will be in the front of the list
        self.assertEqual(['hv1', 'hv3', 'hv2'], chassis)

    def test_least_loaded_chassis_with_rebalance(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple4']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple4']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        # least loaded chassis will be in the front of the list
        self.assertEqual(['hv2', 'hv1'], chassis)

    def test_least_loaded_chassis_per_priority(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple5']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple5']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        # we should now have the following hv's per priority:
        # p5: hv2 (since it currently does not have p5 ports)
        # p4: hv3 or hv4 (since both currently do not have p4 ports)
        # p3: hv5 (since it currently does not have p3 ports)
        # p2: hv1 (since it currently does not have p2 ports)
        # p1: hv3 or hv4 (since they only have one p1 port;
        #                 cant be hv2 since it was already selected)
        self.assertEqual(chassis[0], 'hv2')
        self.assertIn(chassis[1], ['hv3', 'hv4'])
        self.assertEqual(chassis[2], 'hv5')
        self.assertEqual(chassis[3], 'hv1')
        self.assertIn(chassis[4], ['hv3', 'hv4'])
        self.assertNotEqual(chassis[1], chassis[4])

    def test_least_loaded_chassis_per_priority2(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple6']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple6']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        # we should now have the following hv's per priority:
        # p3: hv2 (since it currently does not have p3 ports)
        # p2: hv1 or hv3 (since both currently do not have p2 ports)
        # p1: hv1 or hv3 (since they only have two p1 ports;
        #                 cant be hv2 since it was already selected)
        self.assertEqual(chassis[0], 'hv2')
        self.assertIn(chassis[1], ['hv1', 'hv3'])
        self.assertIn(chassis[2], ['hv1', 'hv3'])
        self.assertNotEqual(chassis[1], chassis[2])

    def test_existing_chassis_available_for_existing_gateway(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple1']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple1']
        gateway_name = random.choice(list(mapping['Gateways'].keys()))
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertEqual(ovn_const.MAX_GW_CHASSIS, len(chassis))


class OVNGatewayLeastLoadedSchedulerWithAZ(OVNGatewayLeastLoadedScheduler):

    def setUp(self):
        super().setUp()

        self.fake_chassis_and_azs = {
            'None': {},
            'Multiple1': {'hv1': {'az-0'},
                          'hv2': {'az-0'},
                          'hv3': {'az-1', 'az-3'},
                          'hv4': {'az-2'},
                          'hv5': {'az-1', 'az-2'}},
            'Multiple2': {},
            'Multiple3': {},
            'Multiple4': {},
            'Multiple5': {},
            'Multiple6': {}}

    def test_least_loaded_chassis_available_for_new_gateway1(self):
        mapping = self.fake_chassis_gateway_mappings['Multiple1']
        chassis_and_azs = self.fake_chassis_and_azs['Multiple1']
        gateway_name = self.new_gateway_name
        chassis = self.select(mapping, gateway_name, chassis_and_azs,
                              candidates=mapping['Chassis'])
        self.assertCountEqual(chassis, mapping.get('Chassis'))
        # least loaded will be the first one in the list,
        # networking-ovn will assign highest priority to this first element
        #
        # Order of chassis is now changed to take care of spreading over AZs :
        # hv5 is in az-1 and az-2 so it can stay first (priority order)
        # hv4 is in az-2, already in hv5 so it goes to the remaining
        # hv3 is in az-1, already in hv5 but az-3 is not, so hv3 is the next
        # one
        # hv2 is in az-0 which is new, so it's the next
        # After, hv4 and hv1 are remaining ones, stay in original order.
        self.assertEqual(['hv5', 'hv3', 'hv2', 'hv4', 'hv1'], chassis)
